import logging
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional
from banal import as_bool
from os import environ as env
from normality import stringify
from datetime import datetime, timezone
from aiocron import Cron  # type: ignore
import random


def env_get(name: str) -> Optional[str]:
    """Ensure the env returns a string even on Windows (#100)."""
    return stringify(env.get(name))


def env_str(name: str, default: str) -> str:
    """Ensure the env returns a string even on Windows (#100)."""
    value = stringify(env.get(name))
    return default if value is None else value


def env_legacy(new_name: str, old_name: str, default: str) -> str:
    """Transition to a new environment variable name with a warning."""
    if old_name in env:
        msg = f"Environment variable {old_name} is deprecated, use {new_name} instead."
        warnings.warn(msg)
    return env_str(new_name, env_str(old_name, default))


def random_cron() -> str:
    """Randomize the minute of the cron schedule to avoid thundering herd problem."""
    random_minute = str(random.randint(0, 59))
    return f"{random_minute} * * * *"


VERSION = "4.2.0"
AUTHOR = "OpenSanctions"
HOME_PAGE = "https://www.opensanctions.org/"
EMAIL = "info@opensanctions.org"
CONTACT = {"name": AUTHOR, "url": HOME_PAGE, "email": EMAIL}

TITLE = env_str("YENTE_TITLE", "yente")
DESCRIPTION = """
The yente API provides endpoints that help you determine if any of the people or
companies mentioned in your data are subject to international sanctions, known
to be involved in criminal activity, or if they are politically exposed people.

`yente` is the open source basis for the OpenSanctions SaaS API. Its matching
and entity retrieval functionality is identical to the hosted API, but it does
not include functionality for metered accounting of API requests.

In this service, there is support for the following operations:

* A simple text-based search for interactive applications (``/search``),
* A query-by-example endpoint for screening tasks (``/match``),
* Support for getting graph data for a particular entity (``/entities``),
* Support for the OpenRefine Reconciliation API (``/reconcile``).

The API uses JSON for data transfer and does not support authentication or access
control.

Further reading:

* [Self-hosted OpenSanctions](https://www.opensanctions.org/docs/self-hosted/)
* [Install and deployment](https://www.opensanctions.org/docs/yente/)
* Intro to the [entity data model](https://www.opensanctions.org/docs/entities/)
* Tutorial: [Using the matching API to do screening checks](https://www.opensanctions.org/docs/api/matching/)
* [Data dictionary](https://opensanctions.org/reference/)
"""

TAGS: List[Dict[str, Any]] = [
    {
        "name": "Matching",
        "description": "Endpoints for conducting a user-facing entity search or "
        "matching a local data store against the given dataset.",
        "externalDocs": {
            "description": "Data dictionary",
            "url": "https://opensanctions.org/reference/",
        },
    },
    {
        "name": "System information",
        "description": "Service metadata endpoints for health checking and getting "
        "the application metadata to be used in client applications.",
    },
    {
        "name": "Data access",
        "description": "Endpoints for fetching data from the API, either related to "
        "individual entities, or for bulk data access in various forms.",
        "externalDocs": {
            "description": "Data dictionary",
            "url": "https://opensanctions.org/reference/",
        },
    },
    {
        "name": "Reconciliation",
        "description": "The Reconciliation Service provides four separate endpoints "
        "that work in concert to implement the data matching API used by OpenRefine, "
        "Wikidata and several other services and utilities.",
        "externalDocs": {
            "description": "W3C Community API specification",
            "url": "https://reconciliation-api.github.io/specs/latest/",
        },
    },
]

# Check if we're running in the context of unit tests:
TESTING = False

# Turn on debug logging and other development features:
DEBUG = as_bool(env_str("YENTE_DEBUG", "false"))

MANIFEST_DEFAULT_PATH = Path(__file__).parent.parent / "manifests/default.yml"

# Path name for the manifest YAML file:
# see: https://www.opensanctions.org/docs/yente/datasets/
MANIFEST = env_str("YENTE_MANIFEST", MANIFEST_DEFAULT_PATH.as_posix())

# Authentication settings:
DATA_TOKEN = env_get("YENTE_DATA_TOKEN")

CRON: Optional[Cron] = None
CRONTAB = env_str("YENTE_CRONTAB", random_cron())

# Whether to automatically reindex the data in the background of the API process:
AUTO_REINDEX = as_bool(env_str("YENTE_AUTO_REINDEX", "true"))

# Fetch the entire bulk data file before indexing into the search index:
STREAM_LOAD = as_bool(env_str("YENTE_STREAM_LOAD", "true"))
# this would be cached here:
DATA_PATH = Path(env_str("YENTE_DATA_PATH", "/tmp"))

# Set a proxy for outgoing HTTP requests:
HTTP_PROXY = env_str("YENTE_HTTP_PROXY", "")

# Whether to enable delta updates for the data:
DELTA_UPDATES = as_bool(env_str("YENTE_DELTA_UPDATES", "true"))

RESOURCES_PATH = Path(__file__).parent.joinpath("resources")

BASE_SCHEMA = "Thing"
PORT = int(env_str("YENTE_PORT", env_str("PORT", "8000")))
HOST = env_str("YENTE_HOST", env_str("HOST", "0.0.0.0"))
UPDATE_TOKEN = env_str("YENTE_UPDATE_TOKEN", "unsafe-default")
CACHE_HEADERS = {
    "Cache-Control": "public; max-age=3600",
    "X-Robots-Tag": "none",
}

# Matcher defaults:
DEFAULT_ALGORITHM = env_str("YENTE_DEFAULT_ALGORITHM", "logic-v1")
BEST_ALGORITHM = env_str("YENTE_BEST_ALGORITHM", "logic-v1")

# How many results to return per page of search results max:
MAX_PAGE = 500

# How many entities to accept in a /search-type endpoint by default:
DEFAULT_PAGE = 10

# How many entities to accept in a /match batch at most:
MAX_BATCH = int(env_str("YENTE_MAX_BATCH", "100"))
MAX_RESULTS = 9999
MAX_OFFSET = MAX_RESULTS - MAX_PAGE

# How many results to return per /match query by default:
MATCH_PAGE = int(env_str("YENTE_MATCH_PAGE", "5"))

# How many results to return per /match query at most:
MAX_MATCHES = int(env_str("YENTE_MAX_MATCHES", "500"))

# How many candidates to retrieve as a multiplier of the /match limit:
MATCH_CANDIDATES = int(env_str("YENTE_MATCH_CANDIDATES", "10"))

# Whether to run expensive levenshtein queries inside ElasticSearch:
MATCH_FUZZY = as_bool(env_str("YENTE_MATCH_FUZZY", "true"))

# How many match and search queries to run against ES in parallel:
QUERY_CONCURRENCY = int(env_str("YENTE_QUERY_CONCURRENCY", "50"))

# Default scoring threshold for /match results:
SCORE_THRESHOLD = 0.70

# Default cutoff for scores that should not be returned as /match results:
SCORE_CUTOFF = 0.50

# ElasticSearch and OpenSearch settings:
INDEX_TYPE = env_str("YENTE_INDEX_TYPE", "elasticsearch").lower().strip()
if INDEX_TYPE not in ["elasticsearch", "opensearch"]:
    raise ValueError(f"Invalid index type: {INDEX_TYPE}")
_INDEX_URL = "http://localhost:9200"
INDEX_URL = env_legacy("YENTE_INDEX_URL", "YENTE_ELASTICSEARCH_URL", _INDEX_URL)

_INDEX_USERNAME = env_legacy("YENTE_INDEX_USERNAME", "YENTE_ELASTICSEARCH_USERNAME", "")
INDEX_USERNAME = None if _INDEX_USERNAME == "" else _INDEX_USERNAME
_INDEX_PASSWORD = env_legacy("YENTE_INDEX_PASSWORD", "YENTE_ELASTICSEARCH_PASSWORD", "")
INDEX_PASSWORD = None if _INDEX_PASSWORD == "" else _INDEX_PASSWORD
_INDEX_SNIFF = env_legacy("YENTE_INDEX_SNIFF", "YENTE_ELASTICSEARCH_SNIFF", "false")
INDEX_SNIFF = as_bool(_INDEX_SNIFF)
_INDEX_CA_CERT = env_legacy("YENTE_INDEX_CA_PATH", "YENTE_ELASTICSEARCH_CA_PATH", "")
INDEX_CA_CERT = None if _INDEX_CA_CERT == "" else _INDEX_CA_CERT
INDEX_SHARDS = int(env_legacy("YENTE_INDEX_SHARDS", "YENTE_ELASTICSEARCH_SHARDS", "1"))
INDEX_NAME = env_legacy("YENTE_INDEX_NAME", "YENTE_ELASTICSEARCH_INDEX", "yente")
ENTITY_INDEX = f"{INDEX_NAME}-entities"
INDEX_VERSION = env_str("YENTE_INDEX_VERSION", "011")
assert len(INDEX_VERSION) == 3, "Index version must be 3 characters long."

# ElasticSearch-only options:
ES_CLOUD_ID = env_get("YENTE_ELASTICSEARCH_CLOUD_ID")

# OpenSearch-only options:
OPENSEARCH_REGION = env_get("YENTE_OPENSEARCH_REGION")
OPENSEARCH_SERVICE = env_get("YENTE_OPENSEARCH_SERVICE")

# Log output can be formatted as JSON:
LOG_JSON = as_bool(env_str("YENTE_LOG_JSON", "false"))
LOG_LEVEL = logging.DEBUG if DEBUG else logging.INFO

# Used to pad out first_seen, last_seen on static collections
RUN_DT = datetime.now(timezone.utc)
RUN_TIME = RUN_DT.isoformat()[:19]



from banal import as_bool
from normality import slugify
from datetime import datetime
from typing import Dict, Optional, Any
from nomenklatura.dataset import Dataset as NKDataset
from nomenklatura.dataset.util import type_check
from nomenklatura.util import iso_to_version, datetime_iso
from followthemoney.types import registry
from followthemoney.namespace import Namespace
from followthemoney.util import sanitize_text

from yente import settings
from yente.logs import get_logger
from yente.data.util import get_url_local_path

log = get_logger(__name__)
BOOT_TIME = datetime_iso(settings.RUN_DT)


class Dataset(NKDataset):
    def __init__(self, data: Dict[str, Any]):
        name = data["name"]
        norm_name = slugify(name, sep="_")
        if name != norm_name:
            raise ValueError("Invalid dataset name %r (try: %r)" % (name, norm_name))
        super().__init__(data)
        self.load = as_bool(data.get("load"), not self.is_collection)
        self.entities_url = self._get_entities_url(data)
        if self.entities_url is not None:
            entities_path = get_url_local_path(self.entities_url)
            if entities_path is not None:
                self.entities_url = entities_path.as_uri()

        if self.version is None:
            ts = data.get("last_export", BOOT_TIME)
            if self.entities_url is not None:
                path = get_url_local_path(self.entities_url)
                if path is not None and path.exists():
                    mtime = path.stat().st_mtime
                    mdt = datetime.fromtimestamp(mtime)
                    ts = datetime_iso(mdt)
            self.version = iso_to_version(ts) or "static"

        self.delta_url = data.get("delta_url", None)

        namespace = as_bool(data.get("namespace"), False)
        self.ns = Namespace(self.name) if namespace else None
        self.index_version: Optional[str] = None

    def _get_entities_url(self, data: Dict[str, Any]) -> Optional[str]:
        entities_url = sanitize_text(data.get("entities_url", data.get("path")))
        if entities_url is not None:
            return entities_url
        resource_name = type_check(registry.string, data.get("resource_name"))
        resource_type = type_check(registry.string, data.get("resource_type"))
        for resource in self.resources:
            if resource.url is None:
                continue
            if resource_name is not None and resource.name == resource_name:
                return resource.url
            if resource_type is not None and resource.mime_type == resource_type:
                return resource.url
        return None

    def to_dict(self) -> Dict[str, Any]:
        data = super().to_dict()
        data["load"] = self.load
        if self.entities_url:
            data["entities_url"] = self.entities_url
        data["index_version"] = self.index_version
        data["index_current"] = self.index_version == self.version
        if self.ns is not None:
            data["namespace"] = True
        if "children" not in data:
            data["children"] = [c.name for c in self.children]
        return data



from .conftest import client
import re


def test_trace_context() -> None:
    # Works when receiving a valid trace context
    headers = {
        "traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01",
        "tracestate": "rojo=00f067aa0ba902b7",
    }
    parent_pat = re.compile(r"00-0af7651916cd43dd8448eb211c80319c-[0-9a-f]{16}-01")
    state_pat = re.compile(r"yente=[0-9a-f]{16},\s?rojo=00f067aa0ba902b7")
    res = client.get("/search/default?q=vladimir putin", headers=headers)
    assert "traceparent" in res.headers
    assert "tracestate" in res.headers
    assert parent_pat.match(res.headers["traceparent"])
    assert state_pat.match(res.headers["tracestate"])
    # Works when not receiving a trace context
    res = client.get("/search/default?q=vladimir putin")
    assert "traceparent" in res.headers
    assert "tracestate" in res.headers
    assert re.match(
        r"00-[0-9a-f]{32}-[0-9a-f]{16}-[0-9a-f]{2}", res.headers["traceparent"]
    )
    assert re.match(r"yente=[0-9a-f]{16}", res.headers["tracestate"])
    # Works with a broken trace context
    headers = {
        "traceparent": "ff-0af7651916cd43dd8448eb211c80319c-0000000000000000-01",
        "tracestate": "rojo=00f067aa0ba902b7",
    }
    res = client.get("/search/default?q=vladimir putin")
    assert "traceparent" in res.headers
    assert "tracestate" in res.headers
    assert re.match(
        r"00-[0-9a-f]{32}-[0-9a-f]{16}-[0-9a-f]{2}", res.headers["traceparent"]
    )
    assert re.match(r"yente=[0-9a-f]{16}", res.headers["tracestate"])



import httpx
from pathlib import Path
from normality import WS
from urllib.parse import urlparse
from followthemoney.types import registry
from prefixdate.precision import Precision
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Dict, List, Iterable, Optional, Set, Generator
from rigour.text.scripts import is_modern_alphabet
from rigour.text.distance import levenshtein
from rigour.text.phonetics import metaphone
from fingerprints import remove_types, clean_name_light
from nomenklatura.util import fingerprint_name, names_word_list

from yente import settings


def _clean_phonetic(original: str) -> Optional[str]:
    # We're being extra picky what phonemes are put into the search index,
    # so that we can reduce the number of false positives.
    if not is_modern_alphabet(original):
        return None
    return fingerprint_name(original)


def expand_dates(dates: List[str]) -> List[str]:
    """Expand a date into less precise versions of itself."""
    expanded = set(dates)
    for date in dates:
        for prec in (Precision.DAY, Precision.MONTH, Precision.YEAR):
            if len(date) > prec.value:
                expanded.add(date[: prec.value])
    return list(expanded)


def phonetic_names(names: List[str]) -> List[str]:
    """Generate phonetic forms of the given names."""
    phonemes: List[str] = []
    for word in names_word_list(names, normalizer=_clean_phonetic, min_length=2):
        token = metaphone(word)
        if len(token) > 2:
            phonemes.append(token)
    return phonemes


def _name_parts(name: Optional[str]) -> Iterable[str]:
    if name is None:
        return
    for part in name.split(WS):
        if len(part) > 1:
            yield part


def index_name_parts(names: List[str]) -> List[str]:
    """Generate a list of indexable name parts from the given names."""
    parts: List[str] = []
    for name in names:
        fp = fingerprint_name(name)
        parts.extend(_name_parts(fp))
        cleaned = remove_types(name, clean=clean_name_light)
        parts.extend(_name_parts(cleaned))
    return parts


def index_name_keys(names: List[str]) -> List[str]:
    """Generate a indexable name keys from the given names."""
    keys: Set[str] = set()
    for name in names:
        for key in (fingerprint_name(name), clean_name_light(name)):
            if key is not None:
                key = key.replace(" ", "")
                keys.add(key)
    return list(keys)


def pick_names(names: List[str], limit: int = 3) -> List[str]:
    """Try to pick a few non-overlapping names to search for when matching
    an entity. The problem here is that if we receive an API query for an
    entity with hundreds of aliases, it becomes prohibitively expensive to
    search. This function decides which ones should be queried as pars pro
    toto in the index before the Python comparison algo later checks all of
    them.

    This is a bit over the top and will come back to haunt me."""
    if len(names) <= limit:
        return names
    picked: List[str] = []
    fingerprinted_ = [fingerprint_name(n) for n in names]
    names = [n for n in fingerprinted_ if n is not None]

    # Centroid:
    picked_name = registry.name.pick(names)
    if picked_name is not None:
        picked.append(picked_name)

    # Pick the least similar:
    for _ in range(1, limit):
        candidates: Dict[str, int] = {}
        for cand in names:
            if cand in picked:
                continue
            candidates[cand] = 0
            for pick in picked:
                candidates[cand] += levenshtein(pick, cand)

        if not len(candidates):
            break
        pick, _ = sorted(candidates.items(), key=lambda c: c[1], reverse=True)[0]
        picked.append(pick)

    return picked


def get_url_local_path(url: str) -> Optional[Path]:
    """Check if a given URL is local file path."""
    parsed = urlparse(url)
    scheme = parsed.scheme.lower()
    if scheme in ("file", "") and parsed.path != "":
        path = Path(parsed.path).resolve()
        if not path.exists():
            raise RuntimeError("File not found: %s" % path)
        return path
    return None


class Authenticator(httpx.Auth):
    def auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        if settings.DATA_TOKEN is not None:
            request.headers["Authentication"] = f"Token {settings.DATA_TOKEN}"
        yield request


@asynccontextmanager
async def httpx_session() -> AsyncGenerator[httpx.AsyncClient, None]:
    transport = httpx.AsyncHTTPTransport(retries=3)
    proxy = settings.HTTP_PROXY if settings.HTTP_PROXY != "" else None
    headers = {"User-Agent": f"Yente/{settings.VERSION}"}
    async with httpx.AsyncClient(
        transport=transport,
        http2=True,
        timeout=None,
        proxy=proxy,
        headers=headers,
        auth=Authenticator(),
    ) as client:
        yield client



from typing import Type
from fastapi import Path, Query
from fastapi import HTTPException
from nomenklatura.matching import ALGORITHMS, ScoringAlgorithm, get_algorithm

from yente import settings
from yente.data import get_catalog
from yente.data.dataset import Dataset


PATH_DATASET = Path(
    description="Data source or collection name to be queries",
    examples=["default"],
)
QUERY_PREFIX = Query("", min_length=0, description="Search prefix")
TS_PATTERN = r"^\d{4}-\d{2}-\d{2}(T\d{2}(:\d{2}(:\d{2})?)?)?$"
ALGO_LIST = ", ".join([a.NAME for a in ALGORITHMS])
ALGO_HELP = (
    f"Scoring algorithm to use, options: {ALGO_LIST} (best: {settings.BEST_ALGORITHM})"
)


def get_algorithm_by_name(name: str) -> Type[ScoringAlgorithm]:
    """Return the scoring algorithm class with the given name."""
    name = name.lower().strip()
    if name == "best":
        name = settings.BEST_ALGORITHM
    algorithm = get_algorithm(name)
    if algorithm is None:
        raise HTTPException(400, detail=f"Invalid algorithm: {name}")
    return algorithm


async def get_dataset(name: str) -> Dataset:
    catalog = await get_catalog()
    dataset = catalog.get(name)
    if dataset is None:
        raise HTTPException(404, detail="No such dataset.")
    return dataset



import json
from normality import ascii_text

from .conftest import client


def test_reconcile_metadata():
    resp = client.get("/reconcile/default")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    url = "https://www.opensanctions.org"
    assert data["identifierSpace"].startswith(url), data
    assert len(data["defaultTypes"]) > 3, data
    assert "suggest" in data, data
    assert "extend" in data, data


def test_reconcile_post_query():
    queries = {"mutti": {"query": "Yevgeny Popov"}}
    resp = client.post("/reconcile/default", data={"queries": json.dumps(queries)})
    assert resp.status_code == 200, resp.text
    data = resp.json()
    res = data["mutti"]["result"]
    assert res[0]["id"] == "Q18634850", res


def test_reconcile_post_extend():
    query = {"ids": ["Q7747"], "properties": [{"id": "name"}, {"id": "birthDate"}]}
    resp = client.post("/reconcile/default", data={"extend": json.dumps(query)})
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert len(data["meta"]) == 2
    assert data["meta"][0]["id"] == "name", data["meta"]
    assert "Q7747" in data["rows"], data
    assert "name" in data["rows"]["Q7747"], data
    names = data["rows"]["Q7747"]["name"]
    assert len(names) > 0, names
    assert "putin" in "".join([n["str"] for n in names]).lower(), names


def test_reconcile_invalid():
    queries = {"mutti": {"type": "Banana"}}
    resp = client.post("/reconcile/default", data={"queries": json.dumps(queries)})
    assert resp.status_code == 400, resp.text


def test_reconcile_suggest_entity_no_prefix():
    resp = client.get("/reconcile/default/suggest/entity")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    assert len(data["result"]) == 0, data


def test_reconcile_suggest_entity_prefix():
    resp = client.get("/reconcile/default/suggest/entity?prefix=vladimir%20put")
    assert resp.status_code == 200, resp.text
    res = resp.json()["result"]
    assert len(res) > 0, res
    assert "Q7747" == res[0]["id"], res
    name = ascii_text(res[0]["name"])
    assert name is not None, res
    assert "vladimir" in name.lower(), name


def test_reconcile_suggest_entity_prefix_dummy():
    resp = client.get("/reconcile/default/suggest/entity?prefix=banana%20man")
    assert resp.status_code == 200, resp.text
    res = resp.json()["result"]
    assert len(res) == 0, res


def test_reconcile_suggest_property_no_prefix():
    resp = client.get("/reconcile/default/suggest/property")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    assert len(data["result"]) == 0, data


def test_reconcile_suggest_property_prefix():
    resp = client.get("/reconcile/default/suggest/property?prefix=country")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    res = data["result"]
    assert len(res) > 1, data
    types = [r["id"] for r in res]
    assert "LegalEntity:mainCountry" in types, types
    assert "Thing:country" in types, types


def test_reconcile_suggest_property_prefix_dummy():
    resp = client.get("/reconcile/default/suggest/property?prefix=banana")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    res = data["result"]
    assert len(res) == 0, data


def test_reconcile_suggest_type_no_prefix():
    resp = client.get("/reconcile/default/suggest/type")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    assert len(data["result"]) == 0, data


def test_reconcile_suggest_type_prefix():
    resp = client.get("/reconcile/default/suggest/type?prefix=organ")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    res = data["result"]
    assert len(res) == 1, data
    assert res[0]["id"] == "Organization", data


def test_reconcile_suggest_type_prefix_dummy():
    resp = client.get("/reconcile/default/suggest/type?prefix=banana")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "result" in data
    res = data["result"]
    assert len(res) == 0, data


def test_reconcile_extend_properties():
    resp = client.get("/reconcile/default/extend/property?limit=5&type=LegalEntity")
    assert resp.status_code == 200, resp.text
    data = resp.json()
    assert "type" in data
    assert data["type"] == "LegalEntity", data
    assert data["limit"] == 5, data
    props = data["properties"]
    assert len(props) == 5
    ids = [p["id"] for p in props]
    assert "name" in ids
    assert "country" in ids


def test_reconcile_extend_properties_invalid_type():
    resp = client.get("/reconcile/default/extend/property?limit=5&type=Banana")
    assert resp.status_code == 400, resp.text



from pprint import pprint  # noqa
from typing import Any, Dict, Generator, List, Tuple, Union, Optional
from followthemoney.schema import Schema
from followthemoney.proxy import EntityProxy
from followthemoney.types import registry

from yente.logs import get_logger
from yente.data.dataset import Dataset
from yente.data.util import pick_names, phonetic_names
from yente.data.util import index_name_parts, index_name_keys
from yente.search.mapping import NAMES_FIELD, NAME_PHONETIC_FIELD
from yente.search.mapping import NAME_PART_FIELD, NAME_KEY_FIELD

log = get_logger(__name__)
FilterDict = Dict[str, Union[bool, str, List[str]]]
Clause = Dict[str, Any]


def filter_query(
    shoulds: List[Clause],
    dataset: Optional[Dataset] = None,
    schema: Optional[Schema] = None,
    filters: FilterDict = {},
    include_dataset: List[str] = [],
    exclude_schema: List[str] = [],
    exclude_dataset: List[str] = [],
    changed_since: Optional[str] = None,
) -> Clause:
    filterqs: List[Clause] = []
    must_not: List[Clause] = []
    datasets: List[str] = include_dataset
    if not len(datasets) and dataset is not None:
        datasets = dataset.dataset_names
    for exclude_ds in exclude_dataset:
        # This is logically a bit more consistent, but doesn't describe the use
        # case of wanting to screen all the entities from datasets X, Y but not Z:
        # must_not.append({"term": {"datasets": exclude_ds}})
        if exclude_ds in datasets:
            datasets.remove(exclude_ds)
    if len(datasets):
        filterqs.append({"terms": {"datasets": datasets}})
    else:
        filterqs.append({"match_none": {}})
    if schema is not None:
        schemata = schema.matchable_schemata
        if not schema.matchable:
            schemata.update(schema.descendants)
        names = [s.name for s in schemata]
        filterqs.append({"terms": {"schema": names}})
    for field, values in filters.items():
        if isinstance(values, (bool, str)):
            filterqs.append({"term": {field: {"value": values}}})
            continue
        values = [v for v in values if len(v)]
        if len(values):
            filterqs.append({"terms": {field: values}})
    if changed_since is not None:
        filterqs.append({"range": {"last_change": {"gt": changed_since}}})

    for schema_name in exclude_schema:
        must_not.append({"term": {"schema": schema_name}})
    return {
        "bool": {
            "filter": filterqs,
            "must_not": must_not,
            "should": shoulds,
            "minimum_should_match": 1,
        }
    }


def names_query(entity: EntityProxy, fuzzy: bool = True) -> List[Clause]:
    names = entity.get_type_values(registry.name, matchable=True)
    names.extend(entity.get("weakAlias", quiet=True))
    shoulds: List[Clause] = []
    for name in pick_names(names, limit=5):
        match = {
            NAMES_FIELD: {
                "query": name,
                "operator": "AND",
                "boost": 3.0,
            }
        }
        if fuzzy:
            match[NAMES_FIELD]["fuzziness"] = "AUTO"
        shoulds.append({"match": match})
    for key in index_name_keys(names):
        term = {NAME_KEY_FIELD: {"value": key, "boost": 4.0}}
        shoulds.append({"term": term})
    for token in set(index_name_parts(names)):
        term = {NAME_PART_FIELD: {"value": token, "boost": 1.0}}
        shoulds.append({"term": term})
    for phoneme in set(phonetic_names(names)):
        term = {NAME_PHONETIC_FIELD: {"value": phoneme, "boost": 0.8}}
        shoulds.append({"term": term})
    return shoulds


def entity_query(
    dataset: Dataset,
    entity: EntityProxy,
    filters: FilterDict = {},
    fuzzy: bool = True,
    include_dataset: List[str] = [],
    exclude_schema: List[str] = [],
    exclude_dataset: List[str] = [],
    changed_since: Optional[str] = None,
) -> Clause:
    shoulds: List[Clause] = names_query(entity, fuzzy=fuzzy)
    for prop, value in entity.itervalues():
        if prop.type == registry.name or not prop.matchable:
            continue
        if prop.type == registry.address:
            query = {"match": {prop.type.group: value}}
            shoulds.append(query)
        elif prop.type.group is not None:
            shoulds.append({"term": {prop.type.group: value}})
        elif fuzzy:
            shoulds.append({"match": {"text": value}})

    return filter_query(
        shoulds,
        filters=filters,
        dataset=dataset,
        schema=entity.schema,
        include_dataset=include_dataset,
        exclude_schema=exclude_schema,
        exclude_dataset=exclude_dataset,
        changed_since=changed_since,
    )


def text_query(
    dataset: Dataset,
    schema: Schema,
    query: str,
    filters: FilterDict = {},
    fuzzy: bool = False,
    simple: bool = False,
    include_dataset: List[str] = [],
    exclude_schema: List[str] = [],
    exclude_dataset: List[str] = [],
    changed_since: Optional[str] = None,
) -> Clause:
    if not len(query.strip()):
        should: Clause = {"match_all": {}}
    elif simple:
        should = {
            "simple_query_string": {
                "query": query,
                "fields": ["names^3", "text"],
                "default_operator": "AND",
                "analyzer": "osa-analyzer",
                "lenient": True,
            }
        }
    else:
        should = {
            "query_string": {
                "query": query,
                "fields": ["names^3", "text"],
                "default_operator": "AND",
                "fuzziness": "AUTO" if fuzzy else 0,
                "analyzer": "osa-analyzer",
                "lenient": True,
            }
        }
        # log.info("Query", should=should)
    return filter_query(
        [should],
        dataset=dataset,
        schema=schema,
        filters=filters,
        include_dataset=include_dataset,
        exclude_schema=exclude_schema,
        exclude_dataset=exclude_dataset,
        changed_since=changed_since,
    )


def prefix_query(
    dataset: Dataset,
    prefix: str,
) -> Clause:
    if not len(prefix.strip()):
        should: Clause = {"match_none": {}}
    else:
        should = {"match_phrase_prefix": {"names": {"query": prefix, "slop": 2}}}
    return filter_query([should], dataset=dataset)


def facet_aggregations(fields: List[str] = []) -> Clause:
    aggs: Clause = {}
    for field in fields:
        aggs[field] = {"terms": {"field": field, "size": 1000}}
    return aggs


def iter_sorts(sorts: List[str]) -> Generator[Tuple[str, str], None, None]:
    for sort in sorts:
        order = "asc"
        if ":" in sort:
            sort, order = sort.rsplit(":", 1)
        if order not in ["asc", "desc"]:
            order = "asc"
        yield sort, order


def parse_sorts(sorts: List[str], default: Optional[str] = "_score") -> List[Any]:
    """Accept sorts of the form: <field>:<order>, e.g. first_seen:desc."""
    objs: List[Any] = []
    for sort, order in iter_sorts(sorts):
        objs.append({sort: {"order": order, "missing": "_last"}})
    if default is not None:
        objs.append(default)
    return objs



from datetime import datetime
from typing import Any, Dict, List, Union, Optional
from pydantic import BaseModel, Field
from nomenklatura.matching.types import MatchingResult, FeatureDocs

from yente import settings
from yente.data.entity import Entity

EntityProperties = Dict[str, List[Union[str, "EntityResponse"]]]


class ErrorResponse(BaseModel):
    detail: str = Field(..., examples=["Detailed error message"])


class EntityResponse(BaseModel):
    id: str = Field(..., examples=["NK-A7z...."])
    caption: str = Field(..., examples=["John Doe"])
    schema_: str = Field(..., examples=["LegalEntity"], alias="schema")
    properties: EntityProperties = Field(..., examples=[{"name": ["John Doe"]}])
    datasets: List[str] = Field([], examples=[["us_ofac_sdn"]])
    referents: List[str] = Field([], examples=[["ofac-1234"]])
    target: bool = Field(False)
    first_seen: Optional[datetime] = Field(None, examples=[settings.RUN_DT])
    last_seen: Optional[datetime] = Field(None, examples=[settings.RUN_DT])
    last_change: Optional[datetime] = Field(None, examples=[settings.RUN_DT])

    @classmethod
    def from_entity(cls, entity: Entity) -> "EntityResponse":
        return cls.model_validate(entity.to_dict())


EntityResponse.model_rebuild()


class ScoredEntityResponse(EntityResponse):
    score: float = 0.99
    features: Dict[str, float]
    match: bool = False

    @classmethod
    def from_entity_result(
        cls, entity: Entity, result: MatchingResult, threshold: float
    ) -> "ScoredEntityResponse":
        data = entity.to_dict()
        data["score"] = result.score
        data["features"] = result.features
        data["match"] = result.score >= threshold
        return cls.model_validate(data)


class StatusResponse(BaseModel):
    status: str = "ok"


class SearchFacetItem(BaseModel):
    name: str = Field(..., examples=["ru"])
    label: str = Field(..., examples=["Russia"])
    count: int = Field(1, examples=[42])


class SearchFacet(BaseModel):
    label: str = Field(..., examples=["Countries"])
    values: List[SearchFacetItem]


class TotalSpec(BaseModel):
    value: int = Field(..., examples=[42])
    relation: str = Field("eq", examples=["eq"])


class ResultsResponse(BaseModel):
    limit: int = Field(..., examples=[20])
    offset: int = Field(0, examples=[0])
    total: TotalSpec


class SearchResponse(ResultsResponse):
    results: List[EntityResponse]
    facets: Dict[str, SearchFacet]


class EntityExample(BaseModel):
    id: Optional[str] = Field(None, examples=["my-entity-id"])
    schema_: str = Field(..., examples=["Person"], alias="schema")
    properties: Dict[str, Union[str, List[Any]]] = Field(
        ..., examples=[{"name": ["John Doe"]}]
    )


class EntityMatchQuery(BaseModel):
    weights: Dict[str, float] = Field({}, examples=[{"name_literal": 0.8}])
    queries: Dict[str, EntityExample]


class EntityMatches(BaseModel):
    status: int = Field(200, examples=[200])
    results: List[ScoredEntityResponse]
    total: TotalSpec
    query: EntityExample


class EntityMatchResponse(BaseModel):
    responses: Dict[str, EntityMatches]
    matcher: FeatureDocs
    limit: int = Field(..., examples=[5])


class DatasetModel(BaseModel):
    name: str
    title: str
    summary: Optional[str] = None
    url: Optional[str] = None
    load: bool
    entities_url: Optional[str] = None
    version: str
    index_version: Optional[str] = None
    index_current: bool = False
    children: List[str]


class DataCatalogModel(BaseModel):
    datasets: List[DatasetModel]
    current: List[str]
    outdated: List[str]
    index_stale: bool = False


class Algorithm(BaseModel):
    name: str
    description: Optional[str] = None
    features: FeatureDocs


class AlgorithmResponse(BaseModel):
    algorithms: List[Algorithm]
    default: str
    best: str



import time
import aiocron  # type: ignore
from typing import AsyncGenerator, Dict, Type, Callable, Any, Coroutine, Union
from contextlib import asynccontextmanager
from pydantic import ValidationError
from fastapi import FastAPI
from fastapi import Request, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from fastapi.responses import JSONResponse
from structlog.contextvars import clear_contextvars, bind_contextvars

from yente import settings
from yente.exc import YenteError
from yente.logs import get_logger
from yente.routers import reconcile, search, match, admin
from yente.data import refresh_catalog
from yente.search.indexer import update_index_threaded
from yente.provider import close_provider
from yente.middleware import TraceContextMiddleware

log = get_logger("yente")
ExceptionHandler = Callable[[Request, Any], Coroutine[Any, Any, Response]]


async def cron_task() -> None:
    await refresh_catalog()
    if settings.AUTO_REINDEX:
        update_index_threaded()


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    log.info(
        "Setting up background refresh",
        crontab=settings.CRONTAB,
        auto_reindex=settings.AUTO_REINDEX,
    )
    settings.CRON = aiocron.crontab(settings.CRONTAB, func=cron_task)
    if settings.AUTO_REINDEX:
        update_index_threaded()
    yield
    await close_provider()


async def request_middleware(
    request: Request, call_next: RequestResponseEndpoint
) -> Response:
    start_time = time.time()
    client_ip = request.client.host if request.client else "127.0.0.1"
    bind_contextvars(
        client_ip=client_ip,
    )
    try:
        response = await call_next(request)
    except Exception as exc:
        log.exception("Exception during request: %s" % type(exc))
        response = JSONResponse(status_code=500, content={"status": "error"})
    time_delta = time.time() - start_time
    log.info(
        str(request.url.path),
        action="request",
        method=request.method,
        path=request.url.path,
        query=request.url.query,
        agent=request.headers.get("user-agent"),
        referer=request.headers.get("referer"),
        code=response.status_code,
        took=time_delta,
    )
    clear_contextvars()
    return response


async def yente_error_handler(req: Request, exc: YenteError) -> Response:
    if exc.status > 499:
        log.exception(f"App error {exc.status}: {exc.detail}")
    return JSONResponse(status_code=exc.status, content={"detail": exc.detail})


async def validation_error_handler(req: Request, exc: ValidationError) -> Response:
    log.warn(f"Validation error: {exc}")
    body = {"detail": exc.title, "errors": exc.errors()}
    return JSONResponse(status_code=400, content=body)


HANDLERS: Dict[Union[Type[Exception], int], ExceptionHandler] = {
    ValidationError: validation_error_handler,
    YenteError: yente_error_handler,
}


def create_app() -> FastAPI:
    app = FastAPI(
        title=settings.TITLE,
        description=settings.DESCRIPTION,
        version=settings.VERSION,
        contact=settings.CONTACT,
        openapi_tags=settings.TAGS,
        exception_handlers=HANDLERS,
        docs_url=None,
        redoc_url=None,
        lifespan=lifespan,
    )
    app.middleware("http")(request_middleware)
    app.add_middleware(TraceContextMiddleware)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    app.add_middleware(GZipMiddleware, minimum_size=500)
    app.include_router(match.router)
    app.include_router(search.router)
    app.include_router(reconcile.router)
    app.include_router(admin.router)
    return app



import pytest
from pathlib import Path

from yente.data import get_catalog
from yente.data.loader import load_json_lines
from yente.data.util import get_url_local_path
from yente.data.util import phonetic_names


@pytest.mark.asyncio
async def test_manifest():
    catalog = await get_catalog()
    assert len(catalog.datasets), catalog.datasets


@pytest.mark.asyncio
async def test_local_dataset():
    catalog = await get_catalog()
    ds = catalog.require("parteispenden")
    assert ds.load
    assert "donations.ijson" in ds.entities_url
    lines = list()
    async for line in load_json_lines(ds.entities_url, "test"):
        lines.append(line)
    assert len(lines) > 10, lines


def test_get_url_local_path():
    out = get_url_local_path("http://banana.com/bla.txt")
    assert out is None
    out = get_url_local_path("https://banana.com/bla.txt")
    assert out is None
    out = get_url_local_path("file:///etc/passwd")
    assert isinstance(out, Path)
    assert "/etc/passwd" in out.as_posix()
    out = get_url_local_path("/etc/passwd")
    assert isinstance(out, Path)
    assert "/etc/passwd" in out.as_posix()
    out = get_url_local_path(__file__)
    assert isinstance(out, Path)

    with pytest.raises(RuntimeError):
        get_url_local_path("/no/such/path.csv")


def test_phonetic_names():
    phonemes = phonetic_names(["Vladimir Putin"])
    assert len(phonemes) == 2
    assert "PTN" in phonemes
    phonemes = phonetic_names(["Влади́мир Влади́мирович ПУ́ТИН"])
    assert len(phonemes) == 3
    assert "PTN" in phonemes
    shortened = phonetic_names(["Влади́мир В. ПУ́ТИН"])
    assert len(shortened) == 2
    phonemes = phonetic_names(["Vladimir Peter Putin"])
    assert len(phonemes) == 3
    phonemes = phonetic_names(["OAO Gazprom"])
    assert len(phonemes) == 1



from typing import Generator, Set
from typing import Any, Dict, List, Optional
from followthemoney import model
from followthemoney.schema import Schema
from followthemoney.types import registry
from nomenklatura.dataset import DataCatalog

from yente import settings
from yente.logs import get_logger
from yente.data.dataset import Dataset
from yente.data.entity import Entity
from yente.data.common import SearchFacet, SearchFacetItem, TotalSpec
from yente.provider import SearchProvider
from yente.util import EntityRedirect

log = get_logger(__name__)
AggType = Dict[str, Dict[str, List[Dict[str, Any]]]]


def result_entity(data: Dict[str, Any]) -> Optional[Entity]:
    source: Optional[Dict[str, Any]] = data.get("_source")
    if source is None or source.get("schema") is None:
        return None
    source["id"] = data.get("_id")
    entity = Entity.from_dict(model, source)
    entity.datasets = set(source["datasets"])
    return entity


def result_total(result: Dict[str, Any]) -> TotalSpec:
    total: Dict[str, Any] = result.get("hits", {}).get("total")
    return TotalSpec(value=total["value"], relation=total["relation"])


def result_entities(response: Dict[str, Any]) -> Generator[Entity, None, None]:
    hits = response.get("hits", {})
    for hit in hits.get("hits", []):
        entity = result_entity(hit)
        if entity is not None:
            yield entity


def result_facets(
    response: Dict[str, Any], catalog: DataCatalog[Dataset]
) -> Dict[str, SearchFacet]:
    facets: Dict[str, SearchFacet] = {}
    aggs: AggType = response.get("aggregations", {})
    for field, agg in aggs.items():
        facet = SearchFacet(label=field, values=[])
        buckets: List[Dict[str, Any]] = agg.get("buckets", [])
        for bucket in buckets:
            key: Optional[str] = bucket.get("key")
            if key is not None:
                key = str(key)
            count: Optional[int] = bucket.get("doc_count")
            if key is None or count is None:
                continue
            value = SearchFacetItem(name=key, label=key, count=count)
            if field == "datasets":
                facet.label = "Data sources"
                value.label = key
                ds = catalog.get(key)
                if ds is not None:
                    value.label = ds.title or key
            if field == "schema":
                facet.label = "Entity types"
                value.label = key
                schema_obj = model.get(key)
                if schema_obj is not None:
                    value.label = schema_obj.plural
            if field in registry.groups:
                type_ = registry.groups[field]
                facet.label = type_.plural
                value.label = type_.caption(key) or value.label
            facet.values.append(value)
        facets[field] = facet
    return facets


async def search_entities(
    provider: SearchProvider,
    query: Dict[str, Any],
    limit: int = 5,
    offset: int = 0,
    aggregations: Optional[Dict[str, Any]] = None,
    sort: List[Any] = [],
) -> Dict[str, Any]:
    return await provider.search(
        index=settings.ENTITY_INDEX,
        query=query,
        size=limit,
        sort=sort,
        from_=offset,
        aggregations=aggregations,
        rank_precise=True,
    )


async def get_entity(provider: SearchProvider, entity_id: str) -> Optional[Entity]:
    query = {
        "bool": {
            "should": [
                {"ids": {"values": [entity_id]}},
                {"term": {"referents": {"value": entity_id}}},
            ],
            "minimum_should_match": 1,
        }
    }
    response = await provider.search(
        index=settings.ENTITY_INDEX,
        query=query,
        size=2,
    )
    hits = response.get("hits", {})
    for hit in hits.get("hits", []):
        if hit.get("_id") != entity_id:
            raise EntityRedirect(hit.get("_id"))
        entity = result_entity(hit)
        if entity is not None:
            return entity
    return None


async def get_matchable_schemata(
    provider: SearchProvider, dataset: Dataset
) -> Set[Schema]:
    """Get the set of schema used in this dataset that are matchable or
    a parent schema to a matchable schema."""
    filter_ = {"terms": {"datasets": dataset.dataset_names}}
    facet = "schemata"
    response = await provider.search(
        index=settings.ENTITY_INDEX,
        query={"bool": {"filter": [filter_]}},
        size=0,
        aggregations={facet: {"terms": {"field": "schema", "size": 1000}}},
    )
    aggs: AggType = response.get("aggregations", {})
    schemata: Set[Schema] = set()
    for bucket in aggs.get(facet, {}).get("buckets", []):
        schema = model.get(bucket["key"])
        if schema is not None and schema.matchable:
            schemata.update(schema.schemata)
    return schemata



from .conftest import client


def test_entity_404():
    res = client.get("/entities/banana")
    assert res.status_code == 404, res


def test_entity_fetch():
    res = client.get("/entities/Q7747")
    assert res.status_code == 200, res
    data = res.json()
    assert data["id"] == "Q7747"
    assert data["schema"] == "Person"
    assert "eu_fsf" in data["datasets"]
    assert len(data["referents"]) > 5, data["referents"]
    assert data["last_change"].startswith("20")
    assert data["first_seen"].startswith("20")
    assert data["last_seen"].startswith("20")

    props = data["properties"]
    assert isinstance(props["birthDate"][0], str), props["birthDate"]

    assert "sanctions" in props
    sanc = props["sanctions"][0]
    assert isinstance(sanc, dict), sanc
    assert sanc["schema"] == "Sanction", sanc

    # for fam in props["familyPerson"]:
    #     famprops = fam["properties"]
    #     assert len(famprops["relative"]) > 0, famprops
    #     for rel in famprops["relative"]:
    #         if isinstance(rel, dict):
    #             assert rel["id"] != data["id"], rel
    #         else:
    #             assert rel != data["id"], rel
    #     for rel in famprops.get("relative", []):
    #         if isinstance(rel, dict):
    #             assert rel["id"] != data["id"], rel
    #         else:
    #             assert rel != data["id"], rel



# mypy: ignore-errors
import pytest
from yente import settings

from yente.exc import YenteIndexError, YenteNotFoundError
from yente.provider import SearchProvider


@pytest.mark.asyncio
async def test_provider_core(search_provider: SearchProvider):
    # Not sure what to test....
    with pytest.raises(YenteNotFoundError):
        fake_index = settings.ENTITY_INDEX + "-doesnt-exist"
        await search_provider.refresh(fake_index)
        await search_provider.check_health(fake_index)

    temp_index = settings.ENTITY_INDEX + "-provider-admin"
    await search_provider.create_index(temp_index)
    await search_provider.refresh(temp_index)
    assert await search_provider.check_health(temp_index) is True
    await search_provider.delete_index(temp_index)


@pytest.mark.asyncio
async def test_index_lifecycle(search_provider: SearchProvider):
    # Given a non-existent index
    # When creating it we should return nothing
    temp_index = settings.ENTITY_INDEX + "-provider-test"
    pre_indices = await search_provider.get_all_indices()
    assert temp_index not in pre_indices
    await search_provider.create_index(temp_index)
    post_indices = await search_provider.get_all_indices()
    assert temp_index in post_indices
    assert len(post_indices) == len(pre_indices) + 1
    # If it already exists we expect no error
    await search_provider.create_index(temp_index)

    with pytest.raises(YenteIndexError):
        await search_provider.create_index(temp_index + "_FAIL")

    await search_provider.refresh(temp_index)

    await search_provider.delete_index(temp_index)
    del_indices = await search_provider.get_all_indices()
    assert temp_index not in del_indices
    assert len(del_indices) == len(pre_indices)


@pytest.mark.asyncio
async def test_alias_management(search_provider: SearchProvider):
    alias = settings.ENTITY_INDEX + "-alias"
    prefix = alias + "-prefix"
    index_v1 = prefix + "-v1"
    index_v2 = prefix + "-v2"
    index_fail = prefix + "-fail"
    await search_provider.create_index(index_v1)
    await search_provider.clone_index(index_v1, index_v2)

    with pytest.raises(YenteIndexError):
        await search_provider.clone_index(index_fail, index_v2)

    assert not await search_provider.exists_index_alias(alias, index_v1)
    assert not await search_provider.exists_index_alias(alias, index_v2)
    assert await search_provider.get_alias_indices(alias) == []

    with pytest.raises(YenteIndexError):
        await search_provider.rollover_index(alias, index_fail, prefix=prefix)
    await search_provider.rollover_index(alias, index_v1, prefix=prefix)
    assert await search_provider.exists_index_alias(alias, index_v1)
    assert not await search_provider.exists_index_alias(alias, index_v2)
    assert await search_provider.get_alias_indices(alias) == [index_v1]

    await search_provider.rollover_index(alias, index_v2, prefix=prefix)
    assert not await search_provider.exists_index_alias(alias, index_v1)
    assert await search_provider.exists_index_alias(alias, index_v2)
    assert await search_provider.get_alias_indices(alias) == [index_v2]

    await search_provider.delete_index(index_v2)
    assert not await search_provider.exists_index_alias(alias, index_v2)
    assert await search_provider.get_alias_indices(alias) == []



from functools import cache
from typing import Tuple
from normality import slugify
import followthemoney

from yente import settings


@cache
def system_version() -> str:
    """Get the current version of the system."""
    parts = [v.rjust(2, "0") for v in followthemoney.__version__.split(".")]
    ftm_version = "".join(parts)[:6]
    return f"{settings.INDEX_VERSION}{ftm_version}-"


def parse_index_name(index: str) -> Tuple[str, str]:
    """
    Parse a given index name.

    Returns:
        dataset_name: str   The name of the dataset the index is based on
        version: str        The version of the index
    """
    # TODO: If we assert that no dashes are allowed in index names we can remove this check.
    if not index.startswith(settings.ENTITY_INDEX):
        raise ValueError("Index created with a different prefix and cannot be parsed.")
    index_end = index[len(settings.ENTITY_INDEX) + 1 :]
    if "-" not in index_end:
        raise ValueError("Index name does not contain a version.")
    dataset, index_version = index_end.split("-", 1)
    sys_version = system_version()
    if not index_version.startswith(sys_version):
        raise ValueError("Index version does not start with the correct prefix.")
    dataset_version = index_version[len(sys_version) :]
    if len(dataset_version) < 1:
        raise ValueError("Index version must be at least one character long.")
    return (dataset, dataset_version)


def construct_index_name(dataset: str, version: str | None = None) -> str:
    """
    Given a dataset and optionally a version construct a properly versioned index name.
    """
    if len(str(dataset)) < 1:
        raise ValueError("Dataset name must be at least one character long.")
    base = f"{settings.ENTITY_INDEX}-{dataset}"
    if version is None:
        return base
    return f"{base}-{construct_index_version(version)}"


def construct_index_version(version: str) -> str:
    """Given a version ID, return a version string with the version prefix."""
    if len(version) < 1:
        raise ValueError("Version must be at least one character long.")
    sys_version = system_version()
    combined = slugify(f"{sys_version}{version}", "-")
    if combined is None or len(combined) < len(sys_version) + 1:
        raise ValueError("Invalid version: %s%s." % (sys_version, version))
    return combined



import yaml
import httpx
import orjson
import asyncio
import aiofiles
from pathlib import Path
from itertools import count
from typing import Any, AsyncGenerator

from yente import settings
from yente.logs import get_logger
from yente.data.util import get_url_local_path, httpx_session

log = get_logger(__name__)


async def load_yaml_url(url: str) -> Any:
    if url.lower().endswith(".json"):
        return await load_json_url(url)
    path = get_url_local_path(url)
    if path is not None:
        async with aiofiles.open(path, "r") as fh:
            data = await fh.read()
    else:
        async with httpx_session() as client:
            resp = await client.get(url)
            data = resp.text
    return yaml.safe_load(data)


async def load_json_url(url: str) -> Any:
    path = get_url_local_path(url)
    if path is not None:
        async with aiofiles.open(path, "rb") as fh:
            data = await fh.read()
    else:
        async with httpx_session() as client:
            resp = await client.get(url)
            resp.raise_for_status()
            data = resp.content
    return orjson.loads(data)


async def fetch_url_to_path(url: str, path: Path) -> None:
    async with httpx_session() as client:
        async with client.stream("GET", url) as resp:
            resp.raise_for_status()
            async with aiofiles.open(path, "wb") as outfh:
                async for chunk in resp.aiter_bytes():
                    await outfh.write(chunk)


async def read_path_lines(path: Path) -> AsyncGenerator[Any, None]:
    async with aiofiles.open(path, "rb") as fh:
        async for line in fh:
            yield orjson.loads(line)


async def stream_http_lines(url: str) -> AsyncGenerator[Any, None]:
    for retry in count():
        try:
            async with httpx_session() as client:
                async with client.stream("GET", url) as resp:
                    resp.raise_for_status()
                    async for line in resp.aiter_lines():
                        yield orjson.loads(line)
                    return
        except httpx.TransportError as exc:
            if retry > 3:
                raise
            await asyncio.sleep(1.0)
            log.error("Streaming index HTTP error: %s, retrying..." % exc)


async def load_json_lines(url: str, base_name: str) -> AsyncGenerator[Any, None]:
    path = get_url_local_path(url)
    if path is not None:
        log.info("Reading local data", url=url, path=path.as_posix())
        async for line in read_path_lines(path):
            yield line

    elif not settings.STREAM_LOAD:
        path = settings.DATA_PATH.joinpath(base_name)
        log.info("Fetching data", url=url, path=path.as_posix())
        try:
            await fetch_url_to_path(url, path)
            async for line in read_path_lines(path):
                yield line
        finally:
            path.unlink(missing_ok=True)
    else:
        log.info("Streaming data", url=url)
        async for line in stream_http_lines(url):
            yield line



from .conftest import client

from yente import settings


def test_healthz():
    res = client.get("/healthz")
    assert res.status_code == 200, res
    assert res.json().get("status") == "ok", res


def test_readyz():
    res = client.get("/readyz")
    assert res.status_code == 200, res
    assert res.json().get("status") == "ok", res


def test_manifest():
    res = client.get("/manifest")
    assert res.status_code == 200, res
    data = res.json()
    assert "datasets" in data
    assert len(data["datasets"]) > 5


def test_algorithms():
    res = client.get("/algorithms")
    assert res.status_code == 200, res
    data = res.json()
    assert "algorithms" in data
    assert len(data["algorithms"]) > 3


def test_catalog():
    res = client.get("/catalog")
    assert res.status_code == 200, res
    data = res.json()
    assert "current" in data
    assert "eu_fsf" in data["current"]
    assert "datasets" in data
    datasets = {d["name"]: d for d in data["datasets"]}
    assert datasets["us_ofac_sdn"]["index_current"] is False
    assert datasets["eu_fsf"]["index_current"] is True
    donations = datasets["parteispenden"]
    assert donations["load"] is True
    assert donations["index_current"] is True
    assert donations["index_version"] == "100"
    assert donations["version"] == "100"


def test_updatez_get():
    res = client.get("/updatez")
    assert res.status_code == 405, res.text


def test_updatez_no_token_configured():
    before = settings.UPDATE_TOKEN
    settings.UPDATE_TOKEN = ""
    res = client.post(f"/updatez?token={before}")
    assert res.status_code == 403, res.text
    settings.UPDATE_TOKEN = before


def test_updatez_no_token():
    res = client.post("/updatez?sync=true")
    assert res.status_code == 403, res.text


def test_updatez_with_token():
    res = client.post(f"/updatez?token={settings.UPDATE_TOKEN}&sync=true")
    assert res.status_code == 200, res.text



from .conftest import client

EXAMPLE = {
    "schema": "Person",
    "properties": {
        "name": ["Vladimir Putin"],
        "birthDate": ["1952"],
        "country": "Russia",
    },
}

ERMAKOV = {
    "properties": {
        "name": [
            "ERMAKOV Valery Nikolaevich",
            "Ermacov Valeryi Nycolaevych",
            "Ermakov Valerij Nikolaevich",
            "Ermakov Valerij Nikolaevič",
            "Ermakov Valerijj Nikolaevich",
            "Ermakov Valeriy Nikolaevich",
            "Ermakov Valery Nykolaevych",
            "Ermakov Valeryi Nykolaevych",
            "Ermakov Valeryy Nikolaevich",
            "Ermakov Valeryy Nykolaevych",
            "Ermakov Valerȳĭ Nȳkolaevȳch",
            "Iermakov Valerii Mykolaiovych",
            "Jermakov Valerij Mikolajovich",
            "Jermakov Valerij Mikolajovič",
            "Jermakov Valerij Mykolajovyč",
            "Yermakov Valerii Mykolaiovych",
            "Yermakov Valerij Mykolajovych",
            "Yermakov Valeriy Mykolayovych",
            "Êrmakov Valerìj Mikolajovič",
            "ЕРМАКОВ Валерий Николаевич",
        ]
    },
    "schema": "Person",
}


def test_match_putin():
    query = {"queries": {"vv": EXAMPLE, "xx": EXAMPLE, "zz": EXAMPLE}}
    resp = client.post("/match/default", json=query)
    assert resp.status_code == 200, resp.text
    data = resp.json()
    res = data["responses"]["vv"]
    assert res["query"]["schema"] == "Person"
    assert res["query"]["properties"]["country"][0] == "ru"
    assert res["total"]["value"] > 0, res["total"]
    res0 = res["results"][0]
    assert res0["id"] == "Q7747", res0


def test_match_putin_name_based_mode():
    query = {"queries": {"vv": EXAMPLE}}
    resp = client.post("/match/default", json=query, params={"algorithm": "neural-net"})
    assert resp.status_code == 400, resp.text

    resp = client.post("/match/default", json=query, params={"algorithm": "name-based"})
    assert resp.status_code == 200, resp.text
    data = resp.json()
    res = data["responses"]["vv"]
    assert res["query"]["schema"] == "Person"
    assert res["query"]["properties"]["country"][0] == "ru"
    assert res["total"]["value"] > 0, res["total"]
    res0 = res["results"][0]
    assert res0["id"] == "Q7747", res0
    assert res0["score"] > 0.70, res0


def test_match_no_schema():
    query = {"queries": {"fail": {"properties": {"name": "Banana"}}}}
    resp = client.post("/match/default", json=query)
    assert resp.status_code == 422, resp.text

    query = {"queries": {"fail": {"schema": "xxx", "properties": {"name": "Banana"}}}}
    resp = client.post("/match/default", json=query)
    assert resp.status_code == 400, resp.text


def test_match_ermakov():
    query = {"queries": {"ermakov": ERMAKOV}}
    resp = client.post("/match/default", json=query)
    assert resp.status_code == 200, resp.text
    results = resp.json()["responses"]["ermakov"]["results"]
    assert len(results) > 0, results

    params = {"fuzzy": "false"}
    resp = client.post("/match/default", json=query, params=params)
    assert resp.status_code == 200, resp.text
    results2 = resp.json()["responses"]["ermakov"]["results"]
    assert len(results) == len(results2), results2


def test_match_exclude_dataset():
    query = {"queries": {"vv": EXAMPLE}}
    params = {"algorithm": "name-based", "exclude_dataset": "eu_fsf"}
    resp = client.post("/match/default", json=query, params=params)
    assert resp.status_code == 200, resp.text
    data = resp.json()
    res = data["responses"]["vv"]
    assert len(res["results"]) == 0, res


def test_match_include_dataset():
    # When querying Putin
    query = {"queries": {"vv": EXAMPLE}}
    # Using only datasets that do not include Putin
    params = {
        "algorithm": "name-based",
        "include_dataset": ["ae_local_terrorists", "mx_governors"],
    }
    resp = client.post("/match/default", json=query, params=params)
    # We should get a succesful response
    assert resp.status_code == 200, resp.text
    data = resp.json()
    res = data["responses"]["vv"]
    # And we should get no matches
    assert len(res["results"]) == 0, res
    # When using a dataset that includes Putin
    params = {
        "algorithm": "name-based",
        "include_dataset": ["eu_fsf", "ae_local_terrorists"],
    }
    resp = client.post("/match/default", json=query, params=params)
    data = resp.json()
    res = data["responses"]["vv"]
    # And we should get matches
    assert len(res["results"]) > 0, res
    # When we exclude the eu_fsf dataset
    params = {
        "algorithm": "name-based",
        "include_dataset": ["eu_fsf", "mx_governors", "ae_local_terrorists"],
        "exclude_dataset": "eu_fsf",
    }
    # We should get no matches
    resp = client.post("/match/default", json=query, params=params)
    data = resp.json()
    res = data["responses"]["vv"]
    assert len(res["results"]) == 0, res


def test_filter_topic():
    query = {"queries": {"vv": EXAMPLE}}
    params = {"algorithm": "name-based", "topics": "crime.cyber"}
    resp = client.post("/match/default", json=query, params=params)
    assert resp.status_code == 200, resp.text
    data = resp.json()
    res = data["responses"]["vv"]
    assert len(res["results"]) == 0, res


def test_id_pass_through():
    body = dict(ERMAKOV)
    body["id"] = "ermakov"
    query = {"queries": {"no1": body}}
    resp = client.post("/match/default", json=query)
    assert resp.status_code == 200, resp.text
    res = resp.json()["responses"]["no1"]
    assert res["query"]["schema"] == "Person"
    assert res["query"]["id"] == "ermakov"



import asyncio
from typing import AsyncIterator
from contextlib import asynccontextmanager

from yente import settings
from yente.logs import get_logger
from yente.provider.elastic import ElasticSearchProvider
from yente.provider.opensearch import OpenSearchProvider
from yente.provider.base import SearchProvider

log = get_logger(__name__)

__all__ = ["with_provider", "get_provider", "close_provider", "SearchProvider"]

PROVIDERS: dict[int, SearchProvider] = {}


def get_id() -> int:
    return id(asyncio.get_event_loop())


async def _create_provider() -> SearchProvider:
    """Create the search provider based on the configured index type."""
    if settings.INDEX_TYPE == "opensearch":
        return await OpenSearchProvider.create()
    else:
        return await ElasticSearchProvider.create()


async def get_provider() -> SearchProvider:
    """Get the search provider for the current event loop, or create it."""
    loop_id = get_id()
    if loop_id not in PROVIDERS:
        PROVIDERS[loop_id] = await _create_provider()
    return PROVIDERS[loop_id]


async def close_provider() -> None:
    loop_id = get_id()
    provider = PROVIDERS.pop(loop_id, None)
    if provider:
        await provider.close()


@asynccontextmanager
async def with_provider() -> AsyncIterator[SearchProvider]:
    provider = await _create_provider()
    try:
        yield provider
    finally:
        await provider.close()



from yente import settings
from yente.logs import get_logger
from yente.provider import SearchProvider
from yente.search.versions import parse_index_name
from yente.data.manifest import Catalog

log = get_logger(__name__)


async def sync_dataset_versions(provider: SearchProvider, catalog: Catalog) -> None:
    for aliased_index in await provider.get_alias_indices(settings.ENTITY_INDEX):
        try:
            dataset_name, version = parse_index_name(aliased_index)
        except ValueError:
            log.warn("Invalid index name: %s" % aliased_index)
            continue
        dataset = catalog.get(dataset_name)
        if dataset is None:
            log.warn("Dataset has index but no metadata: %s" % dataset_name)
            continue
        if version != dataset.version:
            log.info(
                "Dataset %s is outdated" % dataset_name,
                indexed=version,
                available=dataset.version,
            )
        dataset.index_version = version



from typing import Dict, List, Set, Tuple, Union, Optional
from followthemoney.property import Property
from followthemoney.types import registry

from yente import settings
from yente.logs import get_logger
from yente.data.entity import Entity
from yente.data.common import EntityResponse
from yente.provider import SearchProvider
from yente.search.search import result_entities

log = get_logger(__name__)

Value = Union[str, EntityResponse]
Entities = Dict[str, Entity]
Inverted = Dict[str, Set[Tuple[Property, str]]]


def nest_entity(
    entity: Entity, entities: Entities, inverted: Inverted, path: Set[Optional[str]]
) -> EntityResponse:
    props: Dict[str, List[Value]] = {}
    next_path = set([entity.id]).union(path)

    # Find other entities pointing to the one we're processing:
    if entity.id is not None:
        for prop, adj_id in inverted.get(entity.id, {}):
            if adj_id in path or len(path) > 1:
                continue
            adj = entities.get(adj_id)
            if adj is not None:
                nested = nest_entity(adj, entities, inverted, next_path)
                props.setdefault(prop.name, [])
                props[prop.name].append(nested)

    # Expand nested entities:
    for prop in entity.iterprops():
        if prop.type != registry.entity:
            continue
        values: List[Value] = []
        for value in entity.get(prop):
            if value in path:
                continue
            adj = entities.get(value)
            if adj is not None:
                nested = nest_entity(adj, entities, inverted, next_path)
                values.append(nested)
            else:
                values.append(value)
        props[prop.name] = values
        if not len(values):
            props.pop(prop.name)
    serialized = EntityResponse.from_entity(entity)
    serialized.properties.update(props)
    return serialized


async def serialize_entity(
    provider: SearchProvider, root: Entity, nested: bool = False
) -> EntityResponse:
    if not nested or root.id is None:
        return EntityResponse.from_entity(root)
    inverted: Inverted = {}
    reverse = [root.id]

    entities: Entities = {root.id: root}
    next_entities = set(root.get_type_values(registry.entity))

    while True:
        shoulds = []
        if len(reverse):
            shoulds.append({"terms": {"entities": reverse}})

        if len(next_entities):
            shoulds.append({"ids": {"values": list(next_entities)}})

        if not len(shoulds):
            break
        query = {
            "bool": {
                "should": shoulds,
                "minimum_should_match": 1,
                "must_not": [{"ids": {"values": list(entities.keys())}}],
            }
        }

        resp = await provider.search(
            index=settings.ENTITY_INDEX,
            query=query,
            size=settings.MAX_RESULTS,
        )
        reverse = []
        next_entities.clear()
        for adj in result_entities(resp):
            if adj.id is None:
                continue
            entities[adj.id] = adj

            for prop, value in adj.itervalues():
                if prop.type != registry.entity:
                    continue
                if adj.schema.edge and value not in entities:
                    next_entities.add(value)

                inverted.setdefault(value, set())
                if prop.reverse is not None:
                    inverted[value].add((prop.reverse, adj.id))

    return nest_entity(root, entities, inverted, set())



import json
import asyncio
import warnings
from typing import Any, Dict, List, Optional, cast
from typing import AsyncIterator
from elasticsearch import AsyncElasticsearch, ElasticsearchWarning
from elasticsearch.helpers import async_bulk, BulkIndexError
from elasticsearch import ApiError, NotFoundError
from elasticsearch import TransportError, ConnectionError

from yente import settings
from yente.exc import IndexNotReadyError, YenteIndexError, YenteNotFoundError
from yente.logs import get_logger
from yente.search.mapping import make_entity_mapping, INDEX_SETTINGS
from yente.provider.base import SearchProvider, query_semaphore
from yente.middleware.trace_context import get_trace_context

log = get_logger(__name__)
warnings.filterwarnings("ignore", category=ElasticsearchWarning)


class ElasticSearchProvider(SearchProvider):
    @classmethod
    async def create(cls) -> "ElasticSearchProvider":
        """Get elasticsearch connection."""
        kwargs: Dict[str, Any] = dict(
            request_timeout=30,
            retry_on_timeout=True,
            max_retries=10,
        )
        if settings.INDEX_SNIFF:
            kwargs["sniff_on_start"] = True
            kwargs["sniffer_timeout"] = 60
            kwargs["sniff_on_connection_fail"] = True
        if settings.ES_CLOUD_ID:
            log.info("Connecting to Elastic Cloud ID", cloud_id=settings.ES_CLOUD_ID)
            kwargs["cloud_id"] = settings.ES_CLOUD_ID
        else:
            kwargs["hosts"] = [settings.INDEX_URL]
        if settings.INDEX_USERNAME and settings.INDEX_PASSWORD:
            auth = (settings.INDEX_USERNAME, settings.INDEX_PASSWORD)
            kwargs["basic_auth"] = auth
        if settings.INDEX_CA_CERT:
            kwargs["ca_certs"] = settings.INDEX_CA_CERT
        for retry in range(2, 9):
            try:
                es = AsyncElasticsearch(**kwargs)
                es_ = es.options(request_timeout=15)
                await es_.cluster.health(wait_for_status="yellow")
                return ElasticSearchProvider(es)
            except (TransportError, ConnectionError) as exc:
                log.error("Cannot connect to ElasticSearch: %r" % exc)
                await asyncio.sleep(retry**2)

        raise RuntimeError("Could not connect to ElasticSearch.")

    def __init__(self, client: AsyncElasticsearch) -> None:
        self._client = client

    def client(self, **kwargs: Any) -> AsyncElasticsearch:
        """Get the client with the current context."""
        if trace_context := get_trace_context():
            arg_headers = kwargs.get("headers", {})
            headers = arg_headers | (
                dict(
                    traceparent=str(trace_context.traceparent),
                    tracestate=str(trace_context.tracestate),
                )
            )
            kwargs.update(headers=headers)
        return self._client.options(**kwargs)

    async def close(self) -> None:
        await self._client.close()

    async def refresh(self, index: str) -> None:
        """Refresh the index to make changes visible."""
        try:
            await self.client().indices.refresh(index=index)
        except NotFoundError as nfe:
            raise YenteNotFoundError(f"Index {index} does not exist.") from nfe

    async def get_all_indices(self) -> List[str]:
        """Get a list of all indices in the ElasticSearch cluster."""
        indices: Any = await self.client().cat.indices(format="json")
        return [index.get("index") for index in indices]

    async def get_alias_indices(self, alias: str) -> List[str]:
        """Get a list of indices that are aliased to the entity query alias."""
        try:
            resp = await self.client().indices.get_alias(name=alias)
            return list(resp.keys())
        except NotFoundError:
            return []
        except (ApiError, TransportError) as te:
            raise YenteIndexError(f"Could not get alias indices: {te}") from te

    async def rollover_index(self, alias: str, next_index: str, prefix: str) -> None:
        """Remove all existing indices with a given prefix from the alias and
        add the new one."""
        try:
            actions = []
            actions.append({"remove": {"index": f"{prefix}*", "alias": alias}})
            actions.append({"add": {"index": next_index, "alias": alias}})
            await self.client().indices.update_aliases(actions=actions)
        except (ApiError, TransportError) as te:
            raise YenteIndexError(f"Could not rollover index: {te}") from te

    async def clone_index(self, base_version: str, target_version: str) -> None:
        """Create a copy of the index with the given name."""
        if base_version == target_version:
            raise ValueError("Cannot clone an index to itself.")
        try:
            await self.client().indices.put_settings(
                index=base_version,
                settings={"index.blocks.read_only": True},
            )
            await self.delete_index(target_version)
            await self.client().indices.clone(
                index=base_version,
                target=target_version,
                body={
                    "settings": {"index": {"blocks": {"read_only": False}}},
                },
            )
            await self.client().indices.put_settings(
                index=base_version,
                settings={"index.blocks.read_only": False},
            )
            log.info("Cloned index", base=base_version, target=target_version)
        except (ApiError, TransportError) as te:
            msg = f"Could not clone index {base_version} to {target_version}: {te}"
            raise YenteIndexError(msg) from te

    async def create_index(self, index: str) -> None:
        """Create a new index with the given name."""
        log.info("Create index", index=index)
        try:
            await self.client().indices.create(
                index=index,
                mappings=make_entity_mapping(),
                settings=INDEX_SETTINGS,
            )
        except ApiError as exc:
            if exc.error == "resource_already_exists_exception":
                return
            raise YenteIndexError(f"Could not create index: {exc}") from exc

    async def delete_index(self, index: str) -> None:
        """Delete a given index if it exists."""
        try:
            await self.client().indices.delete(index=index)
        except NotFoundError:
            pass
        except (ApiError, TransportError) as te:
            raise YenteIndexError(f"Could not delete index: {te}") from te

    async def exists_index_alias(self, alias: str, index: str) -> bool:
        """Check if an index exists and is linked into the given alias."""
        try:
            exists = await self.client().indices.exists_alias(name=alias, index=index)
            return True if exists.body else False
        except NotFoundError:
            return False
        except (ApiError, TransportError) as te:
            raise YenteIndexError(f"Could not check index alias: {te}") from te

    async def check_health(self, index: str) -> bool:
        try:
            health = await self.client(request_timeout=5).cluster.health(
                index=index, timeout=0
            )
            return health.get("status") in ("yellow", "green")
        except NotFoundError as nfe:
            raise YenteNotFoundError(f"Index {index} does not exist.") from nfe
        except (ApiError, TransportError) as te:
            log.error(f"Search status failure: {te}")
            return False

    async def search(
        self,
        index: str,
        query: Dict[str, Any],
        size: Optional[int] = None,
        from_: Optional[int] = None,
        sort: Optional[List[Any]] = None,
        aggregations: Optional[Dict[str, Any]] = None,
        rank_precise: bool = False,
    ) -> Dict[str, Any]:
        """Search for entities in the index."""

        # This deals with a case in ElasticSearch where the scoring is off when two
        # indices are aliased together and have very different sizes, leading to
        # different term weightings:
        # https://discuss.elastic.co/t/querying-an-alias-throws-off-scoring-completely/351423/4
        search_type = "dfs_query_then_fetch" if rank_precise else None

        try:
            async with query_semaphore:
                response = await self.client().search(
                    index=index,
                    query=query,
                    size=size,
                    from_=from_,
                    sort=sort,
                    aggregations=aggregations,
                    search_type=search_type,
                )
                return cast(Dict[str, Any], response.body)
        except TransportError as te:
            log.warning(
                f"Backend connection error: {te.message}",
                errors=te.errors,
            )
            raise YenteIndexError(f"Could not connect to index: {te.message}") from te
        except ApiError as ae:
            if ae.error == "index_not_found_exception":
                msg = (
                    f"Index {index} does not exist. This may be caused by a misconfiguration,"
                    " or the initial ingestion of data is still ongoing."
                )
                raise IndexNotReadyError(msg) from ae
            if ae.error == "search_phase_execution_exception":
                raise YenteIndexError(f"Search error: {str(ae)}", status=400) from ae
            log.warning(
                f"API error {ae.status_code}: {ae.message}",
                index=index,
                query=json.dumps(query),
            )
            raise YenteIndexError(f"Could not search index: {ae}") from ae
        except (
            KeyboardInterrupt,
            OSError,
            Exception,
            asyncio.TimeoutError,
            asyncio.CancelledError,
        ) as exc:
            msg = f"Error during search: {str(exc)}"
            raise YenteIndexError(msg, status=500) from exc

    async def bulk_index(self, entities: AsyncIterator[Dict[str, Any]]) -> None:
        """Index a list of entities into the search index."""
        try:
            await async_bulk(
                self.client(),
                entities,
                chunk_size=1000,
                yield_ok=False,
                stats_only=True,
            )
        except BulkIndexError as exc:
            raise YenteIndexError(f"Could not index entities: {exc}") from exc



from asyncio import Semaphore
from typing import Any, Dict, List, Optional
from typing import AsyncIterator

from yente import settings

query_semaphore = Semaphore(settings.QUERY_CONCURRENCY)


class SearchProvider(object):
    async def close(self) -> None:
        raise NotImplementedError

    async def refresh(self, index: str) -> None:
        """Refresh the index to make changes visible."""
        raise NotImplementedError

    async def get_all_indices(self) -> List[str]:
        """Get a list of all indices in the ElasticSearch cluster."""
        raise NotImplementedError

    async def get_alias_indices(self, alias: str) -> List[str]:
        """Get a list of indices that are aliased to the entity query alias."""
        raise NotImplementedError

    async def rollover_index(self, alias: str, next_index: str, prefix: str) -> None:
        """Remove all existing indices with a given prefix from the alias and
        add the new one."""
        raise NotImplementedError

    async def clone_index(self, base_version: str, target_version: str) -> None:
        """Create a copy of the index with the given name."""
        raise NotImplementedError

    async def create_index(self, index: str) -> None:
        """Create a new index with the given name."""
        raise NotImplementedError

    async def delete_index(self, index: str) -> None:
        """Delete a given index if it exists."""
        raise NotImplementedError

    async def exists_index_alias(self, alias: str, index: str) -> bool:
        """Check if an index exists and is linked into the given alias."""
        raise NotImplementedError

    async def check_health(self, index: str) -> bool:
        raise NotImplementedError

    async def search(
        self,
        index: str,
        query: Dict[str, Any],
        size: Optional[int] = None,
        from_: Optional[int] = None,
        sort: Optional[List[Any]] = None,
        aggregations: Optional[Dict[str, Any]] = None,
        rank_precise: bool = False,
    ) -> Dict[str, Any]:
        """Search for entities in the index."""
        raise NotImplementedError

    async def bulk_index(self, entities: AsyncIterator[Dict[str, Any]]) -> None:
        """Index a list of entities into the search index."""
        raise NotImplementedError



from typing import Optional


class YenteError(Exception):
    """Base exception for all Yente errors."""

    def __init__(self, detail: str, status: int = 500):
        self.detail = detail
        self.status = status


class YenteIndexError(YenteError):
    """Errors resulting from the search backend being unhappy."""

    STATUS = 500

    def __init__(self, detail: str, status: int = STATUS, index: Optional[str] = None):
        super().__init__(detail, status)
        self.index = index


class IndexNotReadyError(YenteIndexError):
    """Raised when the index is not ready for searching."""

    STATUS = 503


class YenteNotFoundError(YenteIndexError):
    STATUS = 404



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()


setup(
    name="yente",
    version="4.2.0",
    url="https://opensanctions.org/docs/api/",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="OpenSanctions",
    author_email="info@opensanctions.org",
    packages=find_packages(exclude=["examples", "tests"]),
    namespace_packages=[],
    install_requires=[
        "followthemoney==3.7.9",
        "nomenklatura==3.14.0",
        "rigour==0.7.3",
        "asyncstdlib==3.13.0",
        "aiocron==1.8",
        "aiocsv==1.3.2",
        "aiofiles==24.1.0",
        "aiohttp[speedups]==3.11.9",
        "elasticsearch[async]==8.16.0",
        "opensearch-py[async]==2.8.0",
        "boto3>=1.34.144,<1.36.0",
        "fastapi==0.115.5",
        "uvicorn[standard]==0.32.1",
        "httpx[http2]==0.28.0",
        "python-multipart==0.0.19",
        "email-validator==2.2.0",
        "structlog==24.4.0",
        "pyicu==2.14",
        "jellyfish==1.1.0",
        "orjson==3.10.12",
        "text-unidecode==1.3",
        "click==8.1.6",
        "normality==2.5.0",
        "countrynames==1.16.4",
        "fingerprints==1.2.3",
        "pantomime==0.6.1",
        "cryptography==44.0.0",
    ],
    extras_require={
        "dev": [
            "pip>=10.0.0",
            "bump2version",
            "wheel>=0.29.0",
            "ruff>=0.4.0,<1.0.0",
            "twine",
            "mypy",
            "pytest",
            "pytest-cov",
            "pytest-asyncio",
            "pytest-httpx",
            "anyio==4.6.2.post1",
            "flake8>=2.6.0",
            "black",
            "types-aiofiles>=24.0,<25.0",
            "boto3-stubs",
        ],
    },
    entry_points={
        "console_scripts": [
            "yente = yente.cli:cli",
        ],
    },
    zip_safe=False,
)



from typing import Optional, TypedDict, Dict, List, Any
from typing import AsyncGenerator, Tuple

from yente import settings
from yente.logs import get_logger
from yente.data.dataset import Dataset
from yente.data.loader import load_json_url, load_json_lines

log = get_logger(__name__)


class DeltaIndex(TypedDict):
    versions: Dict[str, str]


class EntityOp(TypedDict):
    op: str
    entity: Dict[str, Any]


class DatasetUpdater(object):
    """A helper object for emitting entity operations to transition from one
    loaded dataset version to the next."""

    def __init__(
        self, dataset: Dataset, base_version: Optional[str], force_full: bool = False
    ) -> None:
        self.dataset = dataset
        self.target_version = dataset.version or "static"
        self.base_version = base_version
        self.force_full = force_full
        self.delta_urls: Optional[List[Tuple[str, str]]] = None

    @classmethod
    async def build(
        cls, dataset: Dataset, base_version: Optional[str], force_full: bool = False
    ) -> "DatasetUpdater":
        """Fetch the index of delta files and decide an index building strategy."""
        obj = DatasetUpdater(dataset, base_version, force_full=force_full)
        if force_full:
            return obj
        if dataset.delta_url is None:
            log.debug("No delta updates available for: %r" % dataset.name)
            return obj
        if not settings.DELTA_UPDATES:
            return obj
        if obj.base_version is None or obj.target_version <= obj.base_version:
            return obj

        index: DeltaIndex = await load_json_url(dataset.delta_url)
        versions = index.get("versions", {})
        sorted_versions = sorted(versions.keys())
        if len(sorted_versions) == 0:
            return obj
        # We initially checked if the base_version was in the sorted_versions,
        # but the base_version can be a version that doesn't have a delta (no changes)
        # so we need to check if the base_version is older than the oldest delta version.
        if obj.base_version < min(sorted_versions):
            log.warning(
                "Loaded version of dataset is older than delta window",
                dataset=dataset.name,
                delta_url=dataset.delta_url,
                base_version=obj.base_version,
                target_version=obj.target_version,
                delta_versions=sorted_versions,
            )
            return obj

        obj.delta_urls = []
        for version in sorted_versions:
            if version <= obj.base_version or version > obj.target_version:
                continue
            obj.delta_urls.append((version, versions[version]))

        obj.target_version = max(sorted_versions)
        return obj

    @property
    def is_incremental(self) -> bool:
        """Check if there is sequence of delta entity patches that can be loaded."""
        if self.force_full:
            return False
        if not settings.DELTA_UPDATES:
            return False
        return self.delta_urls is not None

    def needs_update(self) -> bool:
        """Confirm that the dataset needs to be loaded."""
        if not self.dataset.load:
            return False
        if self.dataset.entities_url is None:
            log.warning(
                "Cannot identify resource with FtM entities",
                dataset=self.dataset.name,
            )
            return False
        if self.force_full:
            return True
        if self.target_version is None:
            raise False
        if self.delta_urls is not None and len(self.delta_urls) == 0:
            return False
        if self.base_version is not None and self.target_version <= self.base_version:
            return False
        return True

    async def load(self) -> AsyncGenerator[EntityOp, None]:
        """Generate entity change operations, including payload data."""
        if self.force_full or self.delta_urls is None:
            if self.dataset.entities_url is None:
                raise RuntimeError("No entities for dataset: %s" % self.dataset.name)
            base_name = f"{self.dataset.name}-{self.target_version}"
            async for data in load_json_lines(self.dataset.entities_url, base_name):
                yield {"op": "ADD", "entity": data}
            return

        for version, url in self.delta_urls:
            base_name = f"{self.dataset.name}-delta-{version}"
            async for data in load_json_lines(url, base_name):
                yield data



from datetime import datetime, timedelta

from .conftest import client


def test_search_putin():
    res = client.get("/search/default?q=vladimir putin")
    assert res.status_code == 200, res
    data = res.json()
    assert "results" in data, data
    results = data.get("results")
    assert len(results), results
    putin = results[0]
    assert putin["id"] == "Q7747", results
    assert putin["first_seen"] is not None, putin
    assert putin["first_seen"].startswith("20")
    assert putin["last_seen"] is not None, putin
    assert putin["last_seen"].startswith("20")
    assert "sanctions" not in putin["datasets"]
    assert "default" not in putin["datasets"]


def test_search_no_query():
    res = client.get("/search/default")
    assert res.status_code == 200, res
    results = res.json()["results"]
    assert len(results) > 9, results


def test_search_invalid_query():
    res = client.get("/search/default?q=invalid/query")
    assert res.status_code == 400, res
    res = client.get("/search/default?q=invalid/query&simple=true")
    assert res.status_code == 200, res


def test_search_missing_dataset():
    res = client.get("/search/banana")
    assert res.status_code == 404, res


def test_search_filter_schema_invalid():
    res = client.get("/search/default?q=angela merkel&schema=Banana")
    assert res.status_code == 400, res


def test_search_filter_schema_remove():
    res = client.get("/search/default?q=angela merkel&schema=Vessel")
    assert res.status_code == 200, res
    results = res.json()["results"]
    assert len(results) == 0, results


def test_search_filter_exclude_schema():
    res = client.get("/search/default?q=moscow")
    assert res.status_code == 200, res
    total = res.json()["total"]["value"]
    assert total > 100, total
    res = client.get("/search/default?q=moscow&exclude_schema=Address")
    assert res.status_code == 200, res
    new_total = res.json()["total"]["value"]
    assert new_total < total, new_total


def test_search_filter_exclude_dataset():
    res = client.get("/search/default?q=vladimir putin")
    assert res.status_code == 200, res
    total = res.json()["total"]["value"]
    assert total > 0, total
    res = client.get("/search/default?q=vladimir putin&exclude_dataset=eu_fsf")
    assert res.status_code == 200, res
    new_total = res.json()["total"]["value"]
    assert new_total == 0


def test_search_filter_include_dataset():
    res = client.get("/search/default?q=vladimir putin")
    assert res.status_code == 200, res
    total = res.json()["total"]["value"]
    assert total > 0, total
    # When we include a dataset that does not contain Putin or is not available
    # in the collection we should get no results
    res = client.get("/search/default?q=vladimir putin&include_dataset=mx_senators")
    assert res.status_code == 200, res
    new_total = res.json()["total"]["value"]
    assert new_total == 0
    # When we include a dataset that contains Putin we should get results
    res = client.get("/search/default?q=vladimir putin&include_dataset=eu_fsf")
    new_total = res.json()["total"]["value"]
    assert new_total > 0
    # When using both include and exclude, the exclude should take precedence
    res = client.get(
        "/search/default?q=vladimir putin&include_dataset=eu_fsf&exclude_dataset=eu_fsf"
    )
    new_total = res.json()["total"]["value"]
    assert new_total == 0


def test_search_filter_changed_since():
    ts = datetime.now() + timedelta(days=1)
    tx = ts.isoformat(sep="T", timespec="minutes")
    res = client.get(f"/search/default?q=vladimir putin&changed_since={tx}")
    assert res.status_code == 200, res
    total = res.json()["total"]["value"]
    assert total == 0, total


def test_search_filter_schema_keep():
    res = client.get("/search/default?q=vladimir putin&schema=Person")
    assert res.status_code == 200, res
    results = res.json()["results"]
    assert len(results) > 0, results


def test_search_filter_countries_remove():
    res = client.get("/search/default?q=vladimir putin&countries=ke")
    assert res.status_code == 200, res
    results = res.json()["results"]
    assert len(results) == 0, results


def test_search_facet_datasets_default():
    res = client.get("/search/default")
    assert res.status_code == 200, res
    datasets = res.json()["facets"]["datasets"]
    names = [c["name"] for c in datasets["values"]]
    assert "eu_fsf" in names, names
    assert "parteispenden" not in names, names


def test_search_facet_datasets_spenden():
    res = client.get("/search/parteispenden")
    assert res.status_code == 200, res
    datasets = res.json()["facets"]["datasets"]
    names = [c["name"] for c in datasets["values"]]
    assert "eu_fsf" not in names, names
    assert "parteispenden" in names, names


def test_search_facet_countries():
    res = client.get("/search/default?q=vladimir putin&countries=ru")
    assert res.status_code == 200, res
    countries = res.json()["facets"]["countries"]
    names = [c["name"] for c in countries["values"]]
    assert "ru" in names, names
    assert "ke" not in names, names
    assert "lb" not in names, names


def test_search_facet_topics():
    res = client.get("/search/default?topics=sanction")
    assert res.status_code == 200, res
    sanctioned = res.json()["total"]["value"]
    assert sanctioned > 0, sanctioned

    res = client.get("/search/default")
    assert res.status_code == 200, res
    topics = res.json()["facets"]["topics"]
    names = [c["name"] for c in topics["values"]]
    assert "sanction" in names, names


def test_search_facet_schema():
    res = client.get("/search/default?schema=Address")
    assert res.status_code == 200, res
    addresses = res.json()["total"]["value"]
    assert addresses > 0, addresses

    res = client.get("/search/default?facets=schema")
    assert res.status_code == 200, res
    schemata = res.json()["facets"]["schema"]
    names = [c["name"] for c in schemata["values"]]
    assert "Address" in names, names


def test_search_facet_parameter():
    res = client.get("/search/default")
    assert res.status_code == 200, res
    facets = res.json()["facets"]
    assert len(list(facets.keys())) == 3
    assert "schema" not in facets
    assert "topics" in facets
    assert "datasets" in facets
    assert "countries" in facets

    res = client.get("/search/default?facets=schema&facets=topics")
    assert res.status_code == 200, res
    facets = res.json()["facets"]
    assert len(list(facets.keys())) == 2
    assert "schema" in facets
    assert "topics" in facets
    assert "datasets" not in facets
    assert "countries" not in facets


def test_search_no_targets():
    res = client.get("/search/default?schema=LegalEntity&target=false")
    assert res.status_code == 200, res
    data = res.json()
    assert "results" in data, data
    results = data.get("results")
    assert not len(results), results
    # for res in results:
    #     assert res["target"] == False, res


def test_search_targets():
    res = client.get("/search/default?schema=LegalEntity&target=true")
    assert res.status_code == 200, res
    data = res.json()
    assert "results" in data, data
    results = data.get("results")
    assert len(results), results
    for res in results:
        assert res["target"] is True, res


def test_search_sorted():
    res = client.get("/search/default?sort=first_seen:desc")
    data = res.json()
    assert "results" in data, data
    results = data.get("results")
    prev_seen = None
    for res in results:
        if prev_seen is not None:
            assert res["first_seen"] <= prev_seen, res
        prev_seen = res["first_seen"]


def test_search_putin_scope():
    res = client.get("/search/peps?q=vladimir putin")
    assert res.status_code == 200, res
    data = res.json()
    results = data.get("results")
    assert len(results) == 0, results


def test_search_limit():
    res = client.get("/search/default?limit=0&q=vladimir putin")
    assert res.status_code == 200, res
    data = res.json()
    assert "results" in data, data
    results = data.get("results")
    assert len(results) == 0, results


def test_search_offset():
    res = client.get("/search/default?offset=100&q=vladimir putin")
    assert res.status_code == 200, res
    data = res.json()
    assert "results" in data, data
    results = data.get("results")
    assert len(results) == 0, results
    assert data["offset"] == 100, data["offset"]


def test_search_range_offset():
    res = client.get("/search/default?offset=9999&q=putin")
    assert res.status_code == 422, res


def test_search_range_limit():
    res = client.get("/search/default?limit=10000&q=putin")
    assert res.status_code == 422, res

    res = client.get("/search/default?limit=500&q=putin")
    assert res.status_code == 200, res
    data = res.json()
    assert data["limit"] < 10000, data
    assert data["offset"] == 0, data



import asyncio
import threading
from typing import Any, AsyncGenerator, Dict, List
from followthemoney import model
from followthemoney.exc import FollowTheMoneyException
from followthemoney.types.date import DateType

from yente import settings
from yente.data.manifest import Catalog
from yente.exc import YenteIndexError
from yente.logs import get_logger
from yente.data.entity import Entity
from yente.data.dataset import Dataset
from yente.data import get_catalog
from yente.data.updater import DatasetUpdater
from yente.search.mapping import (
    NAME_PART_FIELD,
    NAME_KEY_FIELD,
    NAMES_FIELD,
    NAME_PHONETIC_FIELD,
)
from yente.provider import SearchProvider, with_provider
from yente.search.versions import parse_index_name
from yente.search.versions import construct_index_name
from yente.data.util import expand_dates, phonetic_names
from yente.data.util import index_name_parts, index_name_keys


log = get_logger(__name__)
lock = threading.Lock()


async def iter_entity_docs(
    updater: DatasetUpdater, index: str
) -> AsyncGenerator[Dict[str, Any], None]:
    dataset = updater.dataset
    datasets = set(dataset.dataset_names)
    idx = 0
    ops: Dict[str, int] = {"ADD": 0, "DEL": 0, "MOD": 0}
    async for data in updater.load():
        if idx % 1000 == 0 and idx > 0:
            log.info("Index: %d entities..." % idx, index=index)
        op_code = data["op"]
        idx += 1
        ops[op_code] += 1
        if op_code == "DEL":
            yield {
                "_op_type": "delete",
                "_index": index,
                "_id": data["entity"]["id"],
            }
            continue

        try:
            entity = Entity.from_dict(model, data["entity"])
            entity.datasets = entity.datasets.intersection(datasets)
            if not len(entity.datasets):
                entity.datasets.add(dataset.name)
            if dataset.ns is not None:
                entity = dataset.ns.apply(entity)

            texts = entity.pop("indexText")
            doc = entity.to_full_dict(matchable=True)
            names: List[str] = doc.get(NAMES_FIELD, [])
            names.extend(entity.get("weakAlias", quiet=True))
            name_parts = index_name_parts(names)
            texts.extend(name_parts)
            doc[NAME_PART_FIELD] = name_parts
            doc[NAME_KEY_FIELD] = index_name_keys(names)
            doc[NAME_PHONETIC_FIELD] = phonetic_names(names)
            doc[DateType.group] = expand_dates(doc.pop(DateType.group, []))
            doc["text"] = texts

            entity_id = doc.pop("id")
            yield {"_index": index, "_id": entity_id, "_source": doc}
        except FollowTheMoneyException as exc:
            log.warning("Invalid entity: %s" % exc, data=data)
    log.info(
        "Indexed %d entities" % idx,
        added=ops["ADD"],
        modified=ops["MOD"],
        deleted=ops["DEL"],
    )


async def get_index_version(provider: SearchProvider, dataset: Dataset) -> str | None:
    """Return the currently indexed version of a given dataset."""
    versions: List[str] = []
    for index in await provider.get_alias_indices(settings.ENTITY_INDEX):
        try:
            ds, version = parse_index_name(index)
            if ds == dataset.name:
                versions.append(version)
        except ValueError:
            pass
    if len(versions) == 0:
        return None
    # Return the oldest version of the index. If multiple versions are linked to the
    # alias, it's a sign that a previous index update failed. So we're erring on the
    # side of caution and returning the oldest version.
    return min(versions)


async def index_entities(
    provider: SearchProvider, dataset: Dataset, force: bool
) -> None:
    """Index entities in a particular dataset, with versioning of the index."""
    alias = settings.ENTITY_INDEX
    base_version = await get_index_version(provider, dataset)
    updater = await DatasetUpdater.build(dataset, base_version, force_full=force)
    if not updater.needs_update():
        if updater.dataset.load:
            log.info("No update needed", dataset=dataset.name, version=base_version)
        return
    log.info(
        "Indexing entities",
        dataset=dataset.name,
        url=dataset.entities_url,
        version=updater.target_version,
        base_version=updater.base_version,
        incremental=updater.is_incremental,
        # delta_urls=updater.delta_urls,
        force=force,
    )
    next_index = construct_index_name(dataset.name, updater.target_version)
    if not force and await provider.exists_index_alias(alias, next_index):
        log.info("Index is up to date.", index=next_index)
        return

    # await es.indices.delete(index=next_index)
    if updater.is_incremental and not force:
        base_index = construct_index_name(dataset.name, updater.base_version)
        await provider.clone_index(base_index, next_index)
    else:
        await provider.create_index(next_index)

    try:
        docs = iter_entity_docs(updater, next_index)
        await provider.bulk_index(docs)
    except (YenteIndexError, Exception) as exc:
        detail = getattr(exc, "detail", str(exc))
        log.exception(
            "Indexing error: %s" % detail,
            dataset=dataset.name,
            index=next_index,
        )
        aliases = await provider.get_alias_indices(alias)
        if next_index not in aliases:
            log.warn("Deleting partial index", index=next_index)
            await provider.delete_index(next_index)
        if updater.is_incremental and not force:
            # This is tricky: try again with a full reindex if the incremental
            # indexing failed
            log.warn("Retrying with full reindex", dataset=dataset.name)
            return await index_entities(provider, dataset, force=True)
        raise exc

    await provider.refresh(index=next_index)
    dataset_prefix = construct_index_name(dataset.name)
    # FIXME: we're not actually deleting old indexes here any more!
    await provider.rollover_index(
        alias,
        next_index,
        prefix=dataset_prefix,
    )
    log.info("Index is now aliased to: %s" % alias, index=next_index)


async def delete_old_indices(provider: SearchProvider, catalog: Catalog) -> None:
    aliased = await provider.get_alias_indices(settings.ENTITY_INDEX)
    for index in await provider.get_all_indices():
        if not index.startswith(settings.ENTITY_INDEX):
            continue
        if index not in aliased:
            log.info("Deleting orphaned index", index=index)
            await provider.delete_index(index)
        try:
            ds_name, _ = parse_index_name(index)
        except ValueError:
            continue
        dataset = catalog.get(ds_name)
        if dataset is None or not dataset.load:
            log.info(
                "Deleting index of non-scope dataset",
                index=index,
                dataset=ds_name,
            )
            await provider.delete_index(index)


async def update_index(force: bool = False) -> None:
    """Reindex all datasets if there is a new version of their data contenst available,
    or create an initial version of the index from scratch."""
    async with with_provider() as provider:
        catalog = await get_catalog()
        log.info("Index update check")
        for dataset in catalog.datasets:
            with lock:
                await index_entities(provider, dataset, force=force)

        await delete_old_indices(provider, catalog)
        log.info("Index update complete.")


def update_index_threaded(force: bool = False) -> None:
    async def update_in_thread() -> None:
        try:
            await update_index(force=force)
        except (Exception, KeyboardInterrupt) as exc:
            log.exception("Index update error: %s" % exc)

    thread = threading.Thread(
        target=asyncio.run,
        args=(update_in_thread(),),
        daemon=True,
    )
    thread.start()
    # asyncio.to_thread(update_index, force=force)



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from typing import Any, Tuple, List
import secrets
from structlog.contextvars import get_contextvars, bind_contextvars

VENDOR_CODE = (
    "yente"  # It's available! https://w3c.github.io/tracestate-ids-registry/#registry
)


class TraceParent:
    __slots__ = ["version", "trace_id", "parent_id", "trace_flags"]

    def __init__(self, version: str, trace_id: str, parent_id: str, trace_flags: str):
        self.version = version
        self.trace_id = trace_id
        self.parent_id = parent_id
        self.trace_flags = trace_flags

    def __str__(self) -> str:
        return f"{self.version}-{self.trace_id}-{self.parent_id}-{self.trace_flags}"

    @classmethod
    def create(cls) -> "TraceParent":
        return cls("00", secrets.token_hex(16), secrets.token_hex(8), "00")

    @classmethod
    def from_str(cls, traceparent: str | None) -> "TraceParent":
        """
        Parse a traceparent header string into a TraceParent object created with a new parent_id.
        """
        if traceparent is None:
            return cls.create()
        parts = traceparent.split("-")
        try:
            version, trace_id, parent_id, trace_flags = parts[:4]
        except Exception:
            raise ValueError(f"Invalid traceparent: {traceparent}")
        if int(version, 16) == 255:
            raise ValueError(f"Unsupported version: {version}")
        for i in trace_id:
            if i != "0":
                break
        else:
            raise ValueError(f"Invalid trace_id: {trace_id}")
        for i in parent_id:
            if i != "0":
                break
        else:
            raise ValueError(f"Invalid parent_id: {parent_id}")

        return cls(version, trace_id, secrets.token_hex(8), trace_flags)


class TraceState:
    __slots__ = ["tracestate"]

    def __init__(self, tracestate: List[Tuple[str, str]] = []):
        self.tracestate = tracestate

    @classmethod
    def create(cls, parent: TraceParent, prev_state: str = "") -> "TraceState":
        spans_out: List[Tuple[str, str]] = []
        for span in prev_state.split(","):
            parts = span.split("=")
            if len(parts) != 2:
                # We are allowed to discard invalid states
                continue
            vendor, value = parts
            if vendor == VENDOR_CODE:
                continue
            spans_out.append((vendor.lower().strip(), value.lower().strip()))
        spans_out.insert(0, (VENDOR_CODE, f"{parent.parent_id}"))
        return cls(spans_out)

    def __str__(self) -> str:
        return ",".join([f"{k}={v}" for k, v in self.tracestate])


class TraceContext:
    __slots__ = ["traceparent", "tracestate"]

    def __init__(self, traceparent: TraceParent, tracestate: TraceState):
        self.traceparent = traceparent
        self.tracestate = tracestate

    def __repr__(self) -> str:
        return str(
            {
                "traceparent": str(self.traceparent),
                "tracestate": str(self.tracestate),
            }
        )


def get_trace_context() -> TraceContext | None:
    vars = get_contextvars()
    if "trace_context" in vars:
        trace_context = vars["trace_context"]
        if isinstance(trace_context, TraceContext):
            return trace_context
    return None


class TraceContextMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: Any) -> Any:
        parent_header = request.headers.get("traceparent")
        try:
            traceparent = TraceParent.from_str(parent_header)
        except Exception:
            traceparent = TraceParent.create()
        state = request.headers.get("tracestate", "")
        try:
            tracestate = TraceState.create(traceparent, state)
        except Exception:
            tracestate = TraceState.create(traceparent, "")
        context = TraceContext(traceparent, tracestate)
        bind_contextvars(trace_context=context)
        resp = await call_next(request)
        resp.headers["traceparent"] = str(traceparent)
        resp.headers["tracestate"] = str(tracestate)
        return resp



import re
import json
import pytest
from .conftest import FIXTURES_PATH
from typing import List, Any

from yente.data import get_catalog, refresh_catalog
from yente.data.updater import DatasetUpdater


@pytest.fixture
def non_mocked_hosts() -> List[str]:
    return ["localhost"]


@pytest.mark.asyncio
async def test_updater(httpx_mock: Any, sanctions_catalog: None) -> None:
    """
    Test getting the delta versions and updating the index, using the data
    mocks in the fixtures directory.
    """
    # Point the entities to our local fixture of 7 entities
    url_pat = re.compile(
        "https:\/\/data\.opensanctions\.org\/datasets\/[\w-]+\/sanctions\/entities\.ftm\.json"
    )
    httpx_mock.add_response(
        200,
        url=url_pat,
        content=(FIXTURES_PATH / "dataset/t1/entities.ftm.json").read_bytes(),
    )
    # The catalog index gets a copy of a real index, as seen at test writing time
    httpx_mock.add_response(
        200,
        url="https://data.opensanctions.org/datasets/latest/index.json",
        content=(FIXTURES_PATH / "dataset/t1/index.json").read_bytes(),
    )
    await refresh_catalog()
    catalog = await get_catalog()
    dataset = catalog.get("sanctions")
    updater = await DatasetUpdater.build(dataset, dataset.version)
    assert not updater.needs_update()

    updater = await DatasetUpdater.build(dataset, None)
    assert not updater.is_incremental
    assert updater.needs_update()
    assert updater.delta_urls is None

    operations = [x async for x in updater.load()]
    assert len(operations) == 7, operations
    for op in operations:
        assert op["op"] == "ADD"

    base_version = dataset.version
    dataset.version = "20240528134729-3iv"
    url = f"https://data.opensanctions.org/artifacts/sanctions/{dataset.version}/delta.json"
    dataset.delta_url = url
    delta_index_path = FIXTURES_PATH / "dataset/t2/delta.json"

    with open(delta_index_path, "r") as f:
        index = json.load(f)
        for version, delta_url in index["versions"].items():
            if version == base_version:
                continue
            httpx_mock.add_response(
                200,
                url=delta_url,
                content=(
                    FIXTURES_PATH / f"dataset/t2/{version}/entities.delta.json"
                ).read_bytes(),
            )

    # Point the index to our fixture containing the new versions
    httpx_mock.add_response(
        200,
        url=url,
        content=delta_index_path.read_bytes(),
    )
    updater = await DatasetUpdater.build(dataset, base_version)
    assert updater.needs_update()
    assert updater.is_incremental
    assert len(updater.delta_urls) == 4

    operations = [x async for x in updater.load()]
    assert len(operations) == 6, operations
    ops = {"ADD": 0, "DEL": 0, "MOD": 0}
    for op in operations:
        ops[op["op"]] += 1

    assert ops["ADD"] == 4
    assert ops["DEL"] == 1
    assert ops["MOD"] == 1
